# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from functools import partial

import pytest

from plopp import Node, View, node


class DataView(View):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data = None

    def notify_view(self, message):
        node_id = message["node_id"]
        self.data = self.graph_nodes[node_id].request_data()


class SimpleView(View):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.value = None

    def notify_view(self, message):
        self.value = message


def test_two_nodes_parent_child():
    a = Node(lambda: 5)
    b = Node(lambda x: x - 2, x=a)
    assert 'x' in b.kwparents
    assert a is b.kwparents['x']
    assert b in a.children
    assert b() == 3


def test_two_nodes_notify_children():
    a = Node(lambda: 5)
    b = Node(lambda x: x - 2, x=a)
    av = SimpleView(a)
    bv = SimpleView(b)
    assert av in a.views
    assert bv in b.views

    msg = 'hello from b'
    b.notify_children(message=msg)
    assert av.value is None
    assert 'node_id' in bv.value
    assert bv.value['message'] == msg

    msg = 'hello from a'
    a.notify_children(message=msg)
    assert 'node_id' in av.value
    assert av.value['message'] == msg
    assert 'node_id' in bv.value
    assert bv.value['message'] == msg


def test_two_children_notify_children():
    a = Node(lambda: 5)
    b = Node(lambda x: x - 2, x=a)
    c = Node(lambda x: x + 2, x=a)
    av = SimpleView(a)
    bv = SimpleView(b)
    cv = SimpleView(c)

    msg = 'hello from b'
    b.notify_children(message=msg)
    assert av.value is None
    assert cv.value is None
    assert 'node_id' in bv.value
    assert bv.value['message'] == msg

    msg = 'hello from a'
    a.notify_children(message=msg)
    assert 'node_id' in av.value
    assert av.value['message'] == msg
    assert 'node_id' in bv.value
    assert bv.value['message'] == msg
    assert 'node_id' in cv.value
    assert cv.value['message'] == msg


def test_two_children_request_data():
    a = Node(lambda: 5)
    b = Node(lambda x: x - 2, x=a)
    c = Node(lambda x: x + 2, x=a)
    av = DataView(a)
    bv = DataView(b)
    cv = DataView(c)

    msg = 'hello from a'
    a.notify_children(message=msg)
    assert av.data == 5
    assert bv.data == 3
    assert cv.data == 7


def test_data_request_is_cached():
    global log
    log = ''

    def log_and_call(n, f, x=None):
        global log
        log += n
        if x is None:
            return f()
        else:
            return f(x)

    a = Node(partial(log_and_call, n='a', f=lambda: 5))
    b = Node(partial(log_and_call, n='b', f=lambda x: x - 2), x=a)
    c = Node(partial(log_and_call, n='c', f=lambda x: x + 2), x=a)
    d = Node(partial(log_and_call, n='d', f=lambda x: x**2), x=c)
    av = DataView(a)  # noqa: F841
    bv = DataView(b)  # noqa: F841
    cv = DataView(c)  # noqa: F841
    dv = DataView(d)  # noqa: F841

    a.notify_children(message='hello from a')
    assert log == 'abcd'  # 'a' should only appear once in the log
    log = ''
    c.notify_children(message='hello from c')
    assert log == 'cd'  # 'c' requests data from 'a' but 'a' is cached so no 'a' in log


def test_remove_node_args():
    a = Node(lambda: 5)
    b = Node(lambda x: x - 2, a)
    c = Node(lambda x: x + 2, a)
    bv = DataView(b)
    cv = DataView(c)
    assert a is b.parents[0]
    assert b in a.children
    assert c in a.children
    assert bv in b.views
    b.remove()
    assert bv not in b.views
    assert b not in a.children
    assert c in a.children
    assert cv in c.views


def test_remove_node_kwargs():
    a = Node(lambda: 5)
    b = Node(lambda x: x - 2, x=a)
    c = Node(lambda x: x + 2, x=a)
    bv = DataView(b)
    cv = DataView(c)
    assert 'x' in b.kwparents
    assert a is b.kwparents['x']
    assert b in a.children
    assert c in a.children
    assert bv in b.views
    b.remove()
    assert bv not in b.views
    assert b not in a.children
    assert c in a.children
    assert cv in c.views


def test_remove_node_with_children_args():
    a = Node(lambda: 5)
    b = Node(lambda x: x - 2, a)
    av = DataView(a)
    a.remove()
    assert b not in a.children
    assert a not in b.parents
    assert a not in b.kwparents.values()
    assert av not in a.views


def test_remove_node_with_children_multiple_args():
    a = Node(lambda: 5)
    b = Node(lambda: 3)
    c = Node(lambda: 11)
    d = Node(lambda *args: sum(args), a, b, c)
    assert d() == 19
    a.remove()
    assert d() == 14
    b.remove()
    assert d() == 11
    c.remove()
    assert d() == 0


def test_remove_node_with_children_kwargs():
    a = Node(lambda: 5)
    b = Node(lambda x: x - 2, x=a)
    av = DataView(a)
    a.remove()
    assert b not in a.children
    assert a not in b.parents
    assert a not in b.kwparents.values()
    assert av not in a.views


def add(x, y):
    return x + y


def test_node_converts_raw_data_to_Node():
    a = Node(5)
    b = Node(add, x=a, y=2)
    assert a() == 5
    assert b.kwparents['x'] is a
    assert 'y' in b.kwparents
    assert b.kwparents['y']() == 2
    assert b() == 7


def test_node_names_from_args():
    n = Node(add, 1, 3)
    assert n.name == 'add(arg_0, arg_1)'


def test_node_names_from_kwargs():
    n = Node(add, x=1, y=3)
    assert n.name == 'add(x, y)'


def test_node_names_from_args_and_kwargs():
    n = Node(add, 4, y=35)
    assert n.name == 'add(arg_0, y)'


def test_node_name_from_raw_input():
    assert Node(5).name == 'Input <int=5>'
    assert Node(1.2).name == 'Input <float=1.2>'
    assert Node('hello').name == "Input <str='hello'>"
    assert Node([1, 2, 3]).name == 'Input <list>'


def test_input_node_value():
    v = 57.2
    a = Node(v)
    b = Node(lambda: v)
    assert a.input_value == v
    assert b.input_value is None


def test_is_input_node():
    a = Node(5)
    b = Node(lambda x: x - 2, x=a)
    assert a.is_input_node
    assert not b.is_input_node


def test_node_operator_add():
    a = Node(5)
    b = Node(2)
    c = a + b
    assert c() == 7


def test_node_operator_radd():
    a = Node(5)
    b = a + 6
    assert b() == 11


def test_node_operator_sub():
    a = Node(5)
    b = Node(2)
    c = a - b
    assert c() == 3


def test_node_operator_rsub():
    a = Node(5)
    b = a - 3
    assert b() == 2


def test_node_operator_mul():
    a = Node(5)
    b = Node(2)
    c = a * b
    assert c() == 10


def test_node_operator_rmul():
    a = Node(5)
    b = a * 6
    assert b() == 30


def test_node_operator_truediv():
    a = Node(5)
    b = Node(2)
    c = a / b
    assert c() == 2.5


def test_node_operator_rtruediv():
    a = Node(5)
    b = a / 2
    assert b() == 2.5


def test_node_decorator_args():
    @node
    def sub(x, y):
        return x - y

    a = Node(5.5)
    b = Node(3.1)
    c = sub(a, b)
    assert c() == 2.4


def test_node_decorator_kwargs():
    @node
    def mult(x, y):
        return x * y

    a = Node(6.0)
    b = Node(4.0)
    c = mult(x=a, y=b)
    assert c() == 24.0


def test_add_parent():
    a = Node(lambda: 5)
    b = Node(lambda x: x - 2)
    assert not a.children
    assert not b.parents
    assert not b.kwparents
    b.add_parents(a)
    assert a in b.parents
    assert b in a.children


def test_add_multiple_parents():
    a = Node(lambda: 5.0)
    b = Node(lambda: 12.0)
    c = Node(lambda: -3.0)
    d = Node(lambda x, y, z: x * y * z)
    d.add_parents(a, b, c)
    assert a in d.parents
    assert b in d.parents
    assert c in d.parents
    assert d in a.children
    assert d in b.children
    assert d in c.children


def test_add_kwparents():
    a = Node(lambda: 5)
    b = Node(lambda time: time * 101.0)
    assert not a.children
    assert not b.parents
    assert not b.kwparents
    b.add_kwparents(time=a)
    assert a is b.kwparents['time']
    assert b in a.children


def test_add_multiple_kwparents():
    a = Node(lambda: 5.0)
    b = Node(lambda: 12.0)
    c = Node(lambda: -3.0)
    d = Node(lambda x, y, z: x * y * z)
    d.add_kwparents(y=a, z=b, x=c)
    assert a is d.kwparents['y']
    assert b is d.kwparents['z']
    assert c is d.kwparents['x']
    assert d in a.children
    assert d in b.children
    assert d in c.children


def test_adding_same_child_twice_raises():
    a = Node(lambda: 5)
    with pytest.raises(ValueError, match="Node .* is already a child in"):
        Node(lambda x, y: x * y - 2, a, a)
    with pytest.raises(ValueError, match="Node .* is already a child in"):
        Node(lambda x, y: x * y - 2, x=a, y=a)


def test_adding_same_parent_twice_raises():
    a = Node(lambda: 5)
    b = Node(lambda x, y: x * y - 2)
    b.add_parents(a)
    with pytest.raises(ValueError, match="Node .* is already a parent in"):
        b.add_parents(a)


def test_adding_same_parent_twice_at_once_raises():
    a = Node(lambda: 5)
    b = Node(lambda x, y: x * y - 2)
    with pytest.raises(ValueError, match="Node .* is already a parent in"):
        b.add_parents(a, a)


def test_adding_same_kwparent_twice_raises():
    a = Node(lambda: 5)
    b = Node(lambda x, y: x * y - 2)
    with pytest.raises(ValueError, match="Node .* is already a child in"):
        b.add_kwparents(x=a, y=a)


def test_adding_same_view_twice_raises():
    a = Node(lambda: 15.0)
    av = SimpleView(a)
    with pytest.raises(ValueError, match="View .* is already a view in"):
        a.add_view(av)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import matplotlib.pyplot as plt
import numpy as np
import scipp as sc

from plopp import Node
from plopp.data.testing import data_array
from plopp.graphics.imageview import ImageView


def test_cbar():
    da = data_array(ndim=2, binedges=True)
    fig = ImageView(Node(da), cbar=False)
    assert fig.canvas.cax is None


def test_update_on_one_mesh_changes_colors_on_second_mesh():
    da1 = data_array(ndim=2)
    da2 = 3.0 * data_array(ndim=2)
    da2.coords['xx'] += sc.scalar(50.0, unit='m')
    a = Node(da1)
    b = Node(da2)
    f = ImageView(a, b)
    old_b_colors = f.artists[b.id]._mesh.get_facecolors()
    a.func = lambda: da1 * 2.1
    a.notify_children('updated a')
    # No change because the update did not change the colorbar limits
    assert np.allclose(old_b_colors, f.artists[b.id]._mesh.get_facecolors())
    a.func = lambda: da1 * 5.0
    a.notify_children('updated a')
    assert not np.allclose(old_b_colors, f.artists[b.id]._mesh.get_facecolors())


def test_with_string_coord():
    strings = ['a', 'b', 'c', 'd', 'e']
    da = sc.DataArray(
        data=sc.array(dims=['y', 'x'], values=np.random.random((5, 5))),
        coords={
            'x': sc.array(dims=['x'], values=strings, unit='s'),
            'y': sc.arange('y', 5.0, unit='m'),
        },
    )
    fig = ImageView(Node(da))
    assert [t.get_text() for t in fig.canvas.ax.get_xticklabels()] == strings


def test_with_strings_as_bin_edges():
    strings = ['a', 'b', 'c', 'd', 'e', 'f']
    da = sc.DataArray(
        data=sc.array(dims=['y', 'x'], values=np.random.random((5, 5))),
        coords={
            'x': sc.array(dims=['x'], values=strings, unit='s'),
            'y': sc.arange('y', 6.0, unit='m'),
        },
    )
    fig = ImageView(Node(da))
    assert [t.get_text() for t in fig.canvas.ax.get_xticklabels()] == strings


def test_with_strings_as_bin_edges_other_coord_is_bin_centers():
    strings = ['a', 'b', 'c', 'd', 'e', 'f']
    da = sc.DataArray(
        data=sc.array(dims=['y', 'x'], values=np.random.random((5, 5))),
        coords={
            'x': sc.array(dims=['x'], values=strings, unit='s'),
            'y': sc.arange('y', 5.0, unit='m'),
        },
    )
    fig = ImageView(Node(da))
    assert [t.get_text() for t in fig.canvas.ax.get_xticklabels()] == strings


def test_kwargs_are_forwarded_to_artist():
    da = data_array(ndim=2)
    fig = ImageView(Node(da), rasterized=True)
    [artist] = fig.artists.values()
    assert artist._mesh.get_rasterized()
    fig = ImageView(Node(da), rasterized=False)
    [artist] = fig.artists.values()
    assert not artist._mesh.get_rasterized()


def test_figsize():
    da = data_array(ndim=2)
    size = (8.1, 8.3)
    fig = ImageView(Node(da), figsize=size)
    assert np.allclose(fig.canvas.fig.get_size_inches(), size)


def test_grid():
    da = data_array(ndim=2)
    fig = ImageView(Node(da), grid=True)
    assert fig.canvas.ax.xaxis.get_gridlines()[0].get_visible()


def test_ax():
    fig, ax = plt.subplots()
    assert len(ax.collections) == 0
    da = data_array(ndim=2)
    _ = ImageView(Node(da), ax=ax)
    assert len(ax.collections) == 1


def test_cax():
    fig, ax = plt.subplots()
    cax = fig.add_axes([0.9, 0.02, 0.05, 0.98])
    assert len(cax.collections) == 0
    da = data_array(ndim=2)
    _ = ImageView(Node(da), ax=ax, cax=cax)
    assert len(cax.collections) > 0



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from collections.abc import Callable
from functools import partial
from typing import Any

import scipp as sc

from ..core import Node, node
from ..core.typing import FigureLike
from .tools import ToggleTool


def is_figure(x):
    from ..backends.matplotlib.interactive import InteractiveFig

    return isinstance(x, InteractiveFig)


class DrawingTool(ToggleTool):
    """
    Interface between Plopp and Mpltoolbox.

    Parameters
    ----------
    figure:
        The figure where the tool will draw things (points, lines, shapes...).
    input_node:
        The node that provides the raw data which is shown in ``figure``.
    tool:
        The Mpltoolbox tool to use (Points, Lines, Rectangles, Ellipses...).
    func:
        The function to be used to make a node whose parents will be the ``input_node``
        and a node yielding the current state of the tool (current position, size).
    destination:
        Where the output from the ``func`` node will be then sent on. This can either
        be a figure, or another graph node.
    get_artist_info:
        A function that returns another function which will convert the properties of
        the artist that produced the event to something (usually a dict) that is usable
        by the ``destination``.
    value:
        Activate the tool upon creation if ``True``.
    **kwargs:
        Additional arguments are forwarded to the ``ToggleTool`` constructor.
    """

    def __init__(
        self,
        figure: FigureLike,
        input_node: Node,
        tool: Any,
        func: Callable,
        destination: FigureLike | Node,
        get_artist_info: Callable,
        value: bool = False,
        **kwargs,
    ):
        super().__init__(callback=self.start_stop, value=value, **kwargs)

        self._figure = figure
        self._input_node = input_node
        self._draw_nodes = {}
        self._output_nodes = {}
        self._func = func
        self._tool = tool(ax=self._figure.ax, autostart=False)
        self._destination = destination
        self._destination_is_fig = is_figure(self._destination)
        self._get_artist_info = get_artist_info
        self._tool.on_create(self.make_node)
        self._tool.on_change(self.update_node)
        self._tool.on_remove(self.remove_node)

    def make_node(self, artist):
        draw_node = Node(self._get_artist_info(artist=artist, figure=self._figure))
        draw_node.name = f'Draw node {len(self._draw_nodes)}'
        nodeid = draw_node.id
        self._draw_nodes[nodeid] = draw_node
        artist.nodeid = nodeid
        output_node = node(self._func)(self._input_node, draw_node)
        output_node.name = f'Output node {len(self._output_nodes)}'
        self._output_nodes[nodeid] = output_node
        if self._destination_is_fig:
            output_node.add_view(self._destination._view)
            self._destination.update({output_node.id: output_node()})
            self._destination.artists[output_node.id].color = (
                artist.color if hasattr(artist, 'color') else artist.edgecolor
            )
        elif isinstance(self._destination, Node):
            self._destination.add_parents(output_node)
            self._destination.notify_children(artist)

    def update_node(self, artist):
        n = self._draw_nodes[artist.nodeid]
        n.func = self._get_artist_info(artist=artist, figure=self._figure)
        n.notify_children(artist)

    def remove_node(self, artist):
        nodeid = artist.nodeid
        draw_node = self._draw_nodes.pop(nodeid)
        output_node = self._output_nodes[nodeid]
        if self._destination_is_fig:
            self._destination.artists[output_node.id].remove()
            del self._destination.artists[output_node.id]
            self._destination.canvas.draw()
        output_node.remove()
        draw_node.remove()

    def start_stop(self):
        """
        Toggle start or stop of the tool.
        """
        if self.value:
            self._tool.start()
        else:
            self._tool.stop()


def _get_points_info(artist, figure):
    """
    Convert the raw (x, y) position of a point to a dict containing the dimensions of
    each axis, and scalar values with units.
    """
    return lambda: {
        'x': {
            'dim': figure.canvas.dims['x'],
            'value': sc.scalar(artist.x, unit=figure.canvas.units['x']),
        },
        'y': {
            'dim': figure.canvas.dims['y'],
            'value': sc.scalar(artist.y, unit=figure.canvas.units['y']),
        },
    }


def _make_points(**kwargs):
    """
    Intermediate function needed for giving to `partial` to avoid making mpltoolbox a
    hard dependency.
    """
    from mpltoolbox import Points

    return Points(**kwargs)


PointsTool = partial(
    DrawingTool,
    tool=partial(_make_points, mec='w'),
    get_artist_info=_get_points_info,
    icon='crosshairs',
)
"""
Tool to add point markers onto a figure.

Parameters
----------
figure:
    The figure where the tool will draw things (points, lines, shapes...).
input_node:
    The node that provides the raw data which is shown in ``figure``.
func:
    The function to be used to make a node whose parents will be the ``input_node``
    and a node yielding the current state of the tool (current position, size).
destination:
    Where the output from the ``func`` node will be then sent on. This can either
    be a figure, or another graph node.
value:
    Activate the tool upon creation if ``True``.
**kwargs:
    Additional arguments are forwarded to the ``ToggleTool`` constructor.
"""



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from ...graphics import BaseFig


class Figure(BaseFig):
    """
    Mixin class for Matplotlib figures
    """

    def __init_figure__(self, View, *args, **kwargs):
        self._view = View(*args, **kwargs)
        self._args = args
        self._kwargs = kwargs

    @property
    def fig(self):
        """
        Get the underlying Matplotlib figure.
        """
        return self._view.canvas.fig

    @property
    def ax(self):
        """
        Get the underlying Matplotlib axes.
        """
        return self._view.canvas.ax

    @property
    def cax(self):
        """
        Get the underlying Matplotlib colorbar axes.
        """
        return self._view.canvas.cax

    def save(self, filename, **kwargs):
        """
        Save the figure to file.
        The default directory for writing the file is the same as the
        directory where the script or notebook is running.

        Parameters
        ----------
        filename:
            Name of the output file. Possible file extensions are ``.jpg``, ``.png``,
            ``.svg``, and ``.pdf``.
        """
        return self._view.canvas.save(filename, **kwargs)

    def __add__(self, other):
        from .tiled import hstack

        return hstack(self, other)

    def __truediv__(self, other):
        from .tiled import vstack

        return vstack(self, other)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from typing import Literal

import matplotlib.pyplot as plt
import numpy as np
import scipp as sc
from matplotlib import dates as mdates
from matplotlib.collections import QuadMesh
from mpl_toolkits.axes_grid1 import make_axes_locatable

from ...core.utils import maybe_variable_to_number, scalar_to_string
from ..common import BoundingBox, axis_bounds
from .utils import fig_to_bytes, is_sphinx_build, make_figure


def _to_floats(x: np.ndarray) -> np.ndarray:
    return mdates.date2num(x) if np.issubdtype(x.dtype, np.datetime64) else x


def _none_if_not_finite(x: float | None) -> float | int | None:
    if x is None:
        return None
    return x if np.isfinite(x) else None


def _cursor_value_to_variable(x: float, dtype: sc.DType, unit: str) -> sc.Variable:
    if dtype == sc.DType.datetime64:
        # Annoying chain of conversion but matplotlib has its own way of converting
        # dates to numbers (number of days since epoch), and num2date returns a python
        # datetime object, while scipp expects a numpy datetime64.
        return sc.scalar(np.datetime64(mdates.num2date(x).replace(tzinfo=None))).to(
            unit=unit
        )
    return sc.scalar(x, unit=unit)


def _cursor_formatter(x: float, dtype: sc.DType, unit: str) -> str:
    if dtype == sc.DType.datetime64:
        return mdates.num2date(x).replace(tzinfo=None).isoformat()
    return scalar_to_string(sc.scalar(x, unit=unit))


class Canvas:
    """
    Matplotlib-based canvas used to render 2D graphics.
    It provides a figure and some axes, as well as functions for controlling the zoom,
    panning, and the scale of the axes.

    Parameters
    ----------
    ax:
        If supplied, use these axes to create the figure. If none are supplied, the
        canvas will create its own axes.
    cax:
        If supplied, use these axes for the colorbar. If none are supplied, and a
        colorbar is required, the canvas will create its own axes.
    figsize:
        The width and height of the figure, in inches.
    title:
        The title to be placed above the figure.
    grid:
        Display the figure grid if ``True``.
    vmin:
        The minimum value for the vertical axis. If a number (without a unit) is
        supplied, it is assumed that the unit is the same as the current vertical axis
        unit.
    vmax:
        The maximum value for the vertical axis. If a number (without a unit) is
        supplied, it is assumed that the unit is the same as the current vertical axis
        unit.
    autoscale:
        The behavior of the axis limits. If ``auto``, the limits automatically
        adjusts every time the data changes. If ``grow``, the limits are allowed to
        grow with time but they do not shrink.
    aspect:
        The aspect ratio for the axes.
    cbar:
        Add axes to host a colorbar if ``True``.
    legend:
        Show legend if ``True``. If ``legend`` is a tuple, it should contain the
        ``(x, y)`` coordinates of the legend's anchor point in axes coordinates.
    """

    def __init__(
        self,
        ax: plt.Axes = None,
        cax: plt.Axes = None,
        figsize: tuple[float, float] | None = None,
        title: str | None = None,
        grid: bool = False,
        vmin: sc.Variable | float = None,
        vmax: sc.Variable | float = None,
        autoscale: Literal['auto', 'grow'] = 'auto',
        aspect: Literal['auto', 'equal'] = 'auto',
        cbar: bool = False,
        legend: bool | tuple[float, float] = True,
        **ignored,
    ):
        # Note on the `**ignored`` keyword arguments: the figure which owns the canvas
        # creates both the canvas and an artist object (Line or Image). The figure
        # accepts keyword arguments, and has to somehow forward them to the canvas and
        # the artist. Since the figure has no detailed knowledge of the underlying
        # backend that implements the canvas, it cannot have specific arguments (such
        # as `ax` for specifying Matplotlib axes).
        # Instead, we forward all the kwargs from the figure to both the canvas and the
        # artist, and filter out the artist kwargs with `**ignored`.

        self.fig = None
        self.ax = ax
        self.cax = cax
        self._user_vmin = vmin
        self._user_vmax = vmax
        self.units = {}
        self.dims = {}
        self._own_axes = False
        self._autoscale = autoscale
        self._legend = legend

        if self.ax is None:
            self._own_axes = True
            self.fig = make_figure(figsize=(6.0, 4.0) if figsize is None else figsize)
            self.ax = self.fig.add_subplot()
            self.ax.set_aspect(aspect)
            self.ax.grid(grid)
        else:
            self.fig = self.ax.get_figure()
        if self.is_widget():
            self.fig.canvas.toolbar_visible = False
            self.fig.canvas.header_visible = False

        if cbar and (self.cax is None):
            if self.ax.name == 'polar':
                bounds = self.ax.get_position().bounds
                self.cax = self.fig.add_axes(
                    [bounds[0] + bounds[2] + 0.1, 0.1, 0.03, 0.8]
                )
            else:
                divider = make_axes_locatable(self.ax)
                self.cax = divider.append_axes("right", "4%", pad="5%")

        if title:
            self.ax.set_title(title)
        self._coord_formatters = []
        self._bbox = BoundingBox()

    def is_widget(self):
        return hasattr(self.fig.canvas, "on_widget_constructed")

    def to_image(self):
        """
        Convert the underlying Matplotlib figure to an image widget from ``ipywidgets``.
        """
        from ipywidgets import Image

        return Image(value=fig_to_bytes(self.fig), format='png')

    def to_widget(self):
        from ipywidgets import Layout, VBox

        if self.is_widget() and not is_sphinx_build():
            # The Matplotlib canvas tries to fill the entire width of the output cell,
            # which can add unnecessary whitespace between it and other widgets. To
            # prevent this, we wrap the canvas in a VBox, which seems to help.
            widget = VBox([self.fig.canvas])
        else:
            widget = self.to_image()
        # The max_width is set to prevent overflow, see gh-169
        widget.layout = Layout(max_width='80%', overflow='auto')
        return widget

    def autoscale(self):
        """
        Find the limits of the artists on the canvas and adjust the axes ranges.
        Add some padding in the case of 1d lines.
        """
        bbox = BoundingBox()
        lines = [line for line in self.ax.lines if hasattr(line, '_plopp_mask')]
        for line in lines:
            line_mask = sc.array(dims=['x'], values=line._plopp_mask)
            line_x = sc.DataArray(
                data=sc.array(dims=['x'], values=_to_floats(line.get_xdata()))
            )
            line_y = sc.DataArray(
                data=sc.array(dims=['x'], values=_to_floats(line.get_ydata())),
                masks={'mask': line_mask},
            )
            line_bbox = BoundingBox(
                **{**axis_bounds(('xmin', 'xmax'), line_x, self.xscale, pad=True)},
                **{**axis_bounds(('ymin', 'ymax'), line_y, self.yscale, pad=True)},
            )
            bbox = bbox.union(line_bbox)

        for c in self.ax.collections:
            if isinstance(c, QuadMesh):
                coords = c.get_coordinates()
                mesh_x = sc.DataArray(
                    data=sc.array(dims=['x', 'y'], values=coords[..., 0])
                )
                mesh_y = sc.DataArray(
                    data=sc.array(dims=['x', 'y'], values=coords[..., 1])
                )
                mesh_bbox = BoundingBox(
                    **{**axis_bounds(('xmin', 'xmax'), mesh_x, self.xscale)},
                    **{**axis_bounds(('ymin', 'ymax'), mesh_y, self.yscale)},
                )
                bbox = bbox.union(mesh_bbox)

            elif hasattr(c, '_plopp_mask'):
                segments = c.get_segments()
                # Here get_segments() can return empty segments in the case where the
                # data values are NaN, which we filter out.
                lengths = np.array([len(segs) for segs in segments])
                line_mask = sc.array(dims=['x'], values=c._plopp_mask[lengths > 0])
                line_y = sc.DataArray(
                    data=sc.array(
                        dims=['x', 'y'],
                        values=np.array(
                            [
                                s
                                for (s, length) in zip(segments, lengths, strict=True)
                                if length
                            ]
                        )[..., 1],
                    ),
                    masks={'mask': line_mask},
                )
                line_bbox = BoundingBox(
                    **axis_bounds(('ymin', 'ymax'), line_y, self.yscale, pad=True)
                )
                bbox = bbox.union(line_bbox)

        self._bbox = self._bbox.union(bbox) if self._autoscale == 'grow' else bbox
        if self._user_vmin is not None:
            self._bbox.ymin = maybe_variable_to_number(
                self._user_vmin, unit=self.units.get('y')
            )
        if self._user_vmax is not None:
            self._bbox.ymax = maybe_variable_to_number(
                self._user_vmax, unit=self.units.get('y')
            )

        self.ax.set_xlim(
            _none_if_not_finite(self._bbox.xmin), _none_if_not_finite(self._bbox.xmax)
        )
        self.ax.set_ylim(
            _none_if_not_finite(self._bbox.ymin), _none_if_not_finite(self._bbox.ymax)
        )
        self.draw()

    def draw(self):
        """
        Make a draw call to the underlying figure.
        """
        self.fig.canvas.draw_idle()

    def save(self, filename: str, **kwargs):
        """
        Save the figure to file.
        The default directory for writing the file is the same as the
        directory where the script or notebook is running.

        Parameters
        ----------
        filename:
            Name of the output file. Possible file extensions are ``.jpg``, ``.png``,
            ``.svg``, and ``.pdf``.
        """
        self.fig.savefig(filename, **{**{'bbox_inches': 'tight'}, **kwargs})

    def show(self):
        """
        Make a call to Matplotlib's underlying ``show`` function.
        """
        self.fig.show()

    def set_axes(self, dims, units, dtypes):
        """
        Set the axes dimensions and units.

        Parameters
        ----------
        dims:
            The dimensions of the data.
        units:
            The units of the data.
        dtypes:
            The data types of the data.
        """
        self.units = units
        self.dims = dims
        self.dtypes = dtypes
        self._cursor_x_prefix = ''
        self._cursor_y_prefix = ''
        if 'y' in self.dims:
            self._cursor_x_prefix = self.dims['x'] + '='
            self._cursor_y_prefix = self.dims['y'] + '='
        self.ax.format_coord = self.format_coord

    def register_format_coord(self, formatter):
        """
        Register a custom axis formatter for the x-axis.
        """
        self._coord_formatters.append(formatter)

    def format_coord(self, x: float, y: float) -> str:
        """
        Format the coordinates of the mouse pointer to show the value of the
        data at that point.

        Parameters
        ----------
        x:
            The x coordinate of the mouse pointer.
        y:
            The y coordinate of the mouse pointer.
        """
        xstr = _cursor_formatter(x, self.dtypes['x'], self.units['x'])
        ystr = _cursor_formatter(y, self.dtypes['y'], self.units['y'])
        out = f"({self._cursor_x_prefix}{xstr}, {self._cursor_y_prefix}{ystr})"
        if not self._coord_formatters:
            return out
        xpos = (
            self.dims['x'],
            _cursor_value_to_variable(x, self.dtypes['x'], self.units['x']),
        )
        ypos = (
            (
                self.dims['y'],
                _cursor_value_to_variable(y, self.dtypes['y'], self.units['y']),
            )
            if 'y' in self.dims
            else None
        )
        extra = [formatter(xpos, ypos) for formatter in self._coord_formatters]
        extra = [e for e in extra if e is not None]
        if extra:
            out += ": {" + ", ".join(extra) + "}"
        return out

    @property
    def empty(self) -> bool:
        """
        Check if the canvas is empty.
        """
        return not self.dims

    @property
    def title(self) -> str:
        """
        Get or set the title of the plot.
        """
        return self.ax.get_title()

    @title.setter
    def title(self, text: str):
        self.ax.set_title(text)

    @property
    def xlabel(self) -> str:
        """
        Get or set the label of the x-axis.
        """
        return self.ax.get_xlabel()

    @xlabel.setter
    def xlabel(self, lab: str):
        self.ax.set_xlabel(lab)

    @property
    def ylabel(self) -> str:
        """
        Get or set the label of the y-axis.
        """
        return self.ax.get_ylabel()

    @ylabel.setter
    def ylabel(self, lab: str):
        self.ax.set_ylabel(lab)

    @property
    def cblabel(self) -> str:
        """
        Get or set the label of the colorbar.
        """
        return self.cax.get_ylabel()

    @cblabel.setter
    def cblabel(self, lab: str):
        self.cax.set_ylabel(lab)

    @property
    def xscale(self) -> Literal['linear', 'log']:
        """
        Get or set the scale of the x-axis ('linear' or 'log').
        """
        return self.ax.get_xscale()

    @xscale.setter
    def xscale(self, scale: Literal['linear', 'log']):
        self.ax.set_xscale(scale)

    @property
    def yscale(self) -> Literal['linear', 'log']:
        """
        Get or set the scale of the y-axis ('linear' or 'log').
        """
        return self.ax.get_yscale()

    @yscale.setter
    def yscale(self, scale: Literal['linear', 'log']):
        self.ax.set_yscale(scale)

    @property
    def xmin(self) -> float:
        """
        Get or set the lower (left) bound of the x-axis.
        """
        return self.ax.get_xlim()[0]

    @xmin.setter
    def xmin(self, value: float):
        self.ax.set_xlim(value, self.xmax)

    @property
    def xmax(self) -> float:
        """
        Get or set the upper (right) bound of the x-axis.
        """
        return self.ax.get_xlim()[1]

    @xmax.setter
    def xmax(self, value: float):
        self.ax.set_xlim(self.xmin, value)

    @property
    def xrange(self) -> tuple[float, float]:
        """
        Get or set the range/limits of the x-axis.
        """
        return self.ax.get_xlim()

    @xrange.setter
    def xrange(self, value: tuple[float, float]):
        self.ax.set_xlim(value)

    @property
    def ymin(self) -> float:
        """
        Get or set the lower (bottom) bound of the y-axis.
        """
        return self.ax.get_ylim()[0]

    @ymin.setter
    def ymin(self, value: float):
        self.ax.set_ylim(value, self.ymax)

    @property
    def ymax(self) -> float:
        """
        Get or set the upper (top) bound of the y-axis.
        """
        return self.ax.get_ylim()[1]

    @ymax.setter
    def ymax(self, value: float):
        self.ax.set_ylim(self.ymin, value)

    @property
    def yrange(self) -> tuple[float, float]:
        """
        Get or set the range/limits of the y-axis.
        """
        return self.ax.get_ylim()

    @yrange.setter
    def yrange(self, value: tuple[float, float]):
        self.ax.set_ylim(value)

    @property
    def grid(self) -> bool:
        """
        Get or set the visibility of the grid.
        """
        return self.ax.axes.get_xgridlines()[0].get_visible()

    @grid.setter
    def grid(self, visible: bool):
        self.ax.grid(visible)

    def reset_mode(self):
        """
        Reset the Matplotlib toolbar mode to nothing, to disable all Zoom/Pan tools.
        """
        if self.fig.canvas.toolbar.mode == 'zoom rect':
            self.zoom()
        elif self.fig.canvas.toolbar.mode == 'pan/zoom':
            self.pan()

    def zoom(self):
        """
        Activate the underlying Matplotlib zoom tool.
        """
        self.fig.canvas.toolbar.zoom()

    def pan(self):
        """
        Activate the underlying Matplotlib pan tool.
        """
        self.fig.canvas.toolbar.pan()

    def panzoom(self, value: Literal['pan', 'zoom', None]):
        """
        Activate or deactivate the pan or zoom tool, depending on the input value.
        """
        if value == 'zoom':
            self.zoom()
        elif value == 'pan':
            self.pan()
        elif value is None:
            self.reset_mode()

    def download_figure(self):
        """
        Save the figure to a PNG file via a pop-up dialog.
        """
        self.fig.canvas.toolbar.save_figure()

    def logx(self):
        """
        Toggle the scale between ``linear`` and ``log`` along the horizontal axis.
        """
        self.xscale = 'log' if self.xscale == 'linear' else 'linear'
        self._bbox.xmin = np.inf
        self._bbox.xmax = -np.inf
        self.autoscale()

    def logy(self):
        """
        Toggle the scale between ``linear`` and ``log`` along the vertical axis.
        """
        self.yscale = 'log' if self.yscale == 'linear' else 'linear'
        self._bbox.ymin = np.inf
        self._bbox.ymax = -np.inf
        self.autoscale()

    def finalize(self):
        """
        Finalize is called at the end of figure creation. Add any polishing operations
        here: trim the margins around the figure.
        """
        if self._own_axes:
            self.fig.tight_layout()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from html import escape
from itertools import chain

from .node import Node
from .view import View


def _make_graphviz_digraph(*args, **kwargs):
    try:
        from graphviz import Digraph
    except ImportError:
        raise RuntimeError(
            "Failed to import `graphviz`. "
            "Use `pip install graphviz` (requires installed `graphviz` executable) or "
            "`conda install -c conda-forge python-graphviz`."
        ) from None
    return Digraph(*args, **kwargs)


def _walk_graph(start, nodes, edges, views, labels):
    label = (
        escape(str(start.func)) + '\nid = ' + start.id
        if start.name is None
        else escape(start.name)
    )
    nodes[start.id] = label
    for child in start.children:
        if start.id not in edges:
            edges[start.id] = {child.id}
        else:
            edges[start.id].add(child.id)
        _walk_graph(
            start=child,
            nodes=nodes,
            edges=edges,
            views=views,
            labels=labels,
        )
    for arg_name, parent in chain(
        ((f'arg_{i}', p) for i, p in enumerate(start.parents)), start.kwparents.items()
    ):
        key = parent.id
        if key not in labels:
            labels[key] = arg_name
        if key not in nodes:
            if key not in edges:
                edges[key] = {start.id}
            else:
                edges[key].add(start.id)
            _walk_graph(
                start=parent,
                nodes=nodes,
                edges=edges,
                views=views,
                labels=labels,
            )

    for view in start.views:
        need_walk = False
        if view.id not in views:
            need_walk = True
        views[view.id] = view.__class__.__name__
        if start.id not in edges:
            edges[start.id] = {view.id}
        else:
            edges[start.id].add(view.id)
        if need_walk:
            for node in view.graph_nodes.values():
                _walk_graph(
                    start=node,
                    nodes=nodes,
                    edges=edges,
                    views=views,
                    labels=labels,
                )


def _make_graph(dot, nodes, edges, labels, views):
    for key, lab in nodes.items():
        dot.node(key, label=lab)
    for key, lab in views.items():
        dot.node(key, label=lab, shape='ellipse', style='filled', color='lightgrey')
    for parent, children in edges.items():
        for child in children:
            dot.edge(
                parent,
                child,
                label=labels.get(parent, '') if child not in views else '',
            )
    return dot


def show_graph(entry: Node | View, **kwargs):
    """
    Display the connected nodes and views as a graph.

    Parameters
    ----------
    entry:
        An entry point in the graph (node or view). This can be any node/view in the
        graph. The graph will be searched from end to end to construct the diagram.
    **kwargs:
        Additional keyword arguments are forwarded to ``graphviz.Digraph``.

    Returns
    -------
    :
        A visual representation of the graph generated with Graphviz.
    """
    dot = _make_graphviz_digraph(strict=True, graph_attr=kwargs)
    dot.attr('node', shape='box', height='0.1')
    nodes = {}
    edges = {}
    views = {}
    labels = {}
    # If input is a View, get the underlying node
    if hasattr(entry, 'graph_nodes'):
        entry = next(iter(entry.graph_nodes.values()))
    _walk_graph(
        start=entry,
        nodes=nodes,
        edges=edges,
        views=views,
        labels=labels,
    )
    return _make_graph(dot=dot, nodes=nodes, edges=edges, labels=labels, views=views)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from functools import lru_cache

import numpy as np
import scipp as sc

_version = '1'


@lru_cache(maxsize=1)
def _make_pooch():
    import pooch

    return pooch.create(
        path=pooch.os_cache('plopp'),
        env='PLOPP_DATA_DIR',
        base_url='https://public.esss.dk/groups/scipp/plopp/{version}/',
        version=_version,
        registry={'nyc_taxi_data.h5': 'md5:fc0867ec061e4ac0cbe5975a665a0eea'},
    )


_pooch = _make_pooch()


def get_path(name: str) -> str:
    """
    Return the path to a data file bundled with plopp.

    This function only works with example data and cannot handle
    paths to custom files.
    """
    return _pooch.fetch(name)


def nyc_taxi() -> str:
    """
    Return the path to the NYC taxi dataset from 2015 histogrammed in latitude,
    longitude, and hour-of-the-dat, stored in scipp's HDF5 format.

    The original data was found at https://vaex.readthedocs.io/en/latest/datasets.html.

    Attention
    ---------
    This data has been manipulated!
    """
    return get_path('nyc_taxi_data.h5')


def three_bands(npeaks=200, per_peak=500, spread=30.0):
    """
    Generate a 2D dataset with three bands of peaks.

    Parameters
    ----------
    npeaks:
        Number of peaks.
    per_peak:
        Number of points per peak.
    spread:
        Standard deviation (spread or 'width') of the peaks.
    """
    ny = 300
    nx = 300
    rng = np.random.default_rng()
    shape = (npeaks, per_peak)
    x = np.empty(shape)
    y = np.empty(shape)
    xcenters = rng.uniform(0, nx, size=npeaks)
    ycenters = rng.choice([ny / 4, ny / 2, 3 * ny / 4], size=npeaks)
    spreads = rng.uniform(0, spread, size=npeaks)
    for i, (xc, yc, sp) in enumerate(zip(xcenters, ycenters, spreads, strict=True)):
        xy = np.random.normal(loc=(xc, yc), scale=sp, size=[per_peak, 2])
        x[i, :] = xy[:, 0]
        y[i, :] = xy[:, 1]

    xcoord = sc.array(dims=['row'], values=x.ravel(), unit='cm')
    ycoord = sc.array(dims=['row'], values=y.ravel(), unit='cm')
    table = sc.DataArray(
        data=sc.ones(sizes=xcoord.sizes, unit='counts'),
        coords={'x': xcoord, 'y': ycoord},
    )
    return table.hist(y=300, x=300) + sc.scalar(1.0, unit='counts')


def clusters3d(nclusters=100, npercluster=2000):
    """
    Generate a 3D dataset with clusters of points.

    Parameters
    ----------
    nclusters:
        Number of clusters.
    npercluster:
        Number of points per cluster.
    """
    position = np.zeros((nclusters, npercluster, 3))
    values = np.zeros((nclusters, npercluster))

    for n in range(nclusters):
        center = 200.0 * (np.random.random(3) - 0.5)
        r = 10.0 * np.random.normal(size=[npercluster, 3])
        position[n, :] = r + center
        values[n, :] = 1 / np.linalg.norm(r, axis=1) ** 2

    return sc.DataArray(
        data=sc.array(dims=['row'], values=values.flatten()),
        coords={
            'position': sc.vectors(
                dims=['row'],
                unit='m',
                values=position.reshape(nclusters * npercluster, 3),
            )
        },
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from collections.abc import Callable
from html import escape

import ipywidgets as ipw


class Checkboxes(ipw.HBox):
    """
    Widget providing a list of checkboxes, along with a button to toggle them all.

    Parameters
    ----------
    entries:
        List of strings to create the names for the different checkboxes.
    description:
        Global description for all the checkboxes.
    value:
        Default value to set all the checkboxes to.
    """

    def __init__(self, entries: list[str], description: str = "", value: bool = True):
        self.checkboxes = {}
        self._lock = False
        self.description = ipw.Label(value=description)

        for key in entries:
            self.checkboxes[key] = ipw.Checkbox(
                value=value,
                description=f"{escape(key)}",
                indent=False,
                layout={"width": "initial"},
            )

        to_hbox = [
            self.description,
            ipw.HBox(
                list(self.checkboxes.values()), layout=ipw.Layout(flex_flow='row wrap')
            ),
        ]

        if len(self.checkboxes) > 1:
            # Add a master button to control all masks in one go
            self.toggle_all_button = ipw.ToggleButton(
                value=value,
                description="Toggle all",
                disabled=False,
                button_style="",
                layout={"width": "initial"},
            )
            for cbox in self.checkboxes.values():
                ipw.jsdlink((self.toggle_all_button, 'value'), (cbox, 'value'))
            to_hbox.insert(1, self.toggle_all_button)

        super().__init__(to_hbox)

    def _plopp_observe_(self, callback: Callable, **kwargs):
        for chbx in self.checkboxes.values():
            chbx.observe(callback, **kwargs)

    @property
    def value(self) -> dict[str, bool]:
        """
        Returns a dict containing one entry per checkbox, giving the checkbox's value.
        """
        return {key: chbx.value for key, chbx in self.checkboxes.items()}



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from typing import Literal

import scipp as sc

from ..core import Node
from ..core.typing import Plottable
from ..core.utils import coord_as_bin_edges
from ..graphics import imagefigure, linefigure
from .common import preprocess, require_interactive_backend


def _to_bin_edges(da: sc.DataArray, dim: str) -> sc.DataArray:
    """
    Convert dimension coords to bin edges.
    """
    for d in set(da.dims) - {dim}:
        da.coords[d] = coord_as_bin_edges(da, d)
    return da


def _apply_op(da: sc.DataArray, op: str, dim: str) -> sc.DataArray:
    out = getattr(sc, op)(da, dim=dim)
    if out.name:
        out.name = f'{op} of {out.name}'
    return out


def _slice_xy(da: sc.DataArray, xy: dict[str, dict[str, int]]) -> sc.DataArray:
    x = xy['x']
    y = xy['y']
    return da[y['dim'], y['value']][x['dim'], x['value']]


def inspector(
    obj: Plottable,
    dim: str | None = None,
    *,
    operation: Literal['sum', 'mean', 'min', 'max'] = 'sum',
    orientation: Literal['horizontal', 'vertical'] = 'horizontal',
    **kwargs,
):
    """
    Inspector takes in a three-dimensional input and applies a reduction operation
    (``'sum'`` by default) along one of the dimensions specified by ``dim``.
    It displays the result as a two-dimensional image.
    In addition, an 'inspection' tool is available in the toolbar which allows to place
    markers on the image which perform slicing at that position to retain only the third
    dimension and displays the resulting one-dimensional slice on the right hand side
    figure.

    Controls:
    - Click to make new point
    - Drag existing point to move it
    - Middle-click to delete point

    Parameters
    ----------
    obj:
        The object to be plotted.
    dim:
        The dimension along which to apply the reduction operation. This will also be
        the dimension that remains in the one-dimensional slices generated by adding
        markers on the image. If no dim is provided, the last (inner) dim of the input
        data will be used.
    operation:
        The operation to apply along the third (undisplayed) dimension specified by
        ``dim``.
    orientation:
        Display the two panels side-by-side ('horizontal') or one below the other
        ('vertical').
    **kwargs:
        See :py:func:`plopp.plot` for the full list of figure customization arguments.

    Returns
    -------
    :
        A :class:`Box` which will contain two :class:`Figure` and one slider widget.
    """
    require_interactive_backend('inspector')

    in_node = Node(preprocess, obj, ignore_size=True)
    data = in_node()
    if data.ndim != 3:
        raise ValueError(
            'The inspector plot currently only works with '
            f'three-dimensional data, found {data.ndim} dims.'
        )
    if dim is None:
        dim = data.dims[-1]
    bin_edges_node = Node(_to_bin_edges, in_node, dim=dim)
    op_node = Node(_apply_op, da=bin_edges_node, op=operation, dim=dim)
    f2d = imagefigure(op_node, **kwargs)
    f1d = linefigure()

    from ..widgets import Box, PointsTool

    pts = PointsTool(
        figure=f2d,
        input_node=bin_edges_node,
        func=_slice_xy,
        destination=f1d,
        tooltip="Activate inspector tool",
    )
    f2d.toolbar['inspect'] = pts
    out = [f2d, f1d]
    if orientation == 'horizontal':
        out = [out]
    return Box(out)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from copy import copy
from typing import Any

import ipywidgets as ipw
import numpy as np
import scipp as sc

from ...graphics import Camera


class Canvas:
    """
    WebGL canvas used to render 3D graphics.
    It is based on the ``pythreejs`` library. It provides a scene, camera, controls,
    as well as functions for controlling the camera.
    In addition, it draws a RGB axes helper for the XYZ dimensions, and can draw an
    outline box, given a spatial range in the three directions.

    Parameters
    ----------
    figsize:
        The width and height of the renderer in pixels.
    title:
        The title to be placed above the figure. It is possible to use HTML formatting
        to customize the title appearance.
    camera:
        Initial camera configuration (position, target).
    """

    def __init__(
        self,
        figsize: tuple[int, int] = (600, 400),
        title: str | None = None,
        camera: Camera | None = None,
    ):
        import pythreejs as p3

        self.dims = {}
        self.units = {}
        self.outline = None
        self.axticks = None
        self.figsize = np.asarray(figsize)
        self._title_text = title
        self._title = self._make_title()
        width, height = self.figsize
        self._user_camera = Camera() if camera is None else camera
        self.camera = p3.PerspectiveCamera(aspect=width / height)
        self.camera_backup = {}
        self.axes_3d = p3.AxesHelper()
        self.outline = None
        self.scene = p3.Scene(
            children=[self.camera, self.axes_3d], background="#f0f0f0"
        )
        self.controls = p3.OrbitControls(controlling=self.camera)
        self.renderer = p3.Renderer(
            camera=self.camera,
            scene=self.scene,
            controls=[self.controls],
            width=width,
            height=height,
        )

    def to_widget(self):
        # The max_width is set to prevent overflow, see gh-169
        self.renderer.layout = ipw.Layout(max_width='80%', overflow='auto')
        return self.renderer

    def set_axes(self, dims, units, dtypes):
        """
        Set the axes dimensions and units.

        Parameters
        ----------
        dims:
            The dimensions of the data.
        units:
            The units of the data.
        dtypes:
            The data types of the data.
        """
        self.units = units
        self.dims = dims
        self.dtypes = dtypes

    def make_outline(self, limits: tuple[sc.Variable, sc.Variable, sc.Variable]):
        """
        Create an outline box with ticklabels, given a range in the XYZ directions.
        """
        from .outline import Outline

        if self.outline is not None:
            self.remove(self.outline)
        self.outline = Outline(limits=limits)
        self.add(self.outline)
        self._update_camera(limits=limits)
        self.axes_3d.scale = [self.camera.far] * 3

    def _update_camera(self, limits: tuple[sc.Variable, sc.Variable, sc.Variable]):
        """
        Update the camera position when a new object is added to the canvas.
        The camera will look at the mean position of all the objects, and its position
        will be far enough from the center to see all the objects.
        This 'Home' position will be backed up so it can be used when resetting the
        camera via the ``home`` function.
        """
        if not self._user_camera.has_units():
            self._user_camera.set_units(
                self.units.get('x'), self.units.get('y'), self.units.get('z')
            )

        center = [var.mean().value for var in limits]
        distance_fudge_factor = 1.2
        box_size = np.array([(limits[i][1] - limits[i][0]).value for i in range(3)])
        self.camera.position = self._user_camera.get(
            'position', list(np.array(center) + distance_fudge_factor * box_size)
        )
        camera_dist = np.linalg.norm(np.array(self.camera.position) - np.array(center))
        box_mean_size = np.linalg.norm(box_size)
        self.camera.near = self._user_camera.get('near', 0.01 * box_mean_size)
        camera_to_origin = np.linalg.norm(
            np.array(self.camera.position) - np.array([0, 0, 0])
        )
        center_to_origin = np.linalg.norm(np.array(center) - np.array([0, 0, 0]))
        self.camera.far = self._user_camera.get(
            'far',
            5 * max(box_mean_size, camera_dist, camera_to_origin, center_to_origin),
        )
        camera_lookat = tuple(self._user_camera.get('look_at', center))
        self.controls.target = camera_lookat
        self.camera.lookAt(camera_lookat)

        # Save camera settings
        self.camera_backup["reset"] = copy(self.camera.position)
        self.camera_backup["look_at"] = copy(camera_lookat)
        self.camera_backup["center"] = tuple(copy(center))
        self.camera_backup["x_normal"] = [
            center[0] - distance_fudge_factor * box_mean_size,
            center[1],
            center[2],
        ]
        self.camera_backup["y_normal"] = [
            center[0],
            center[1] - distance_fudge_factor * box_mean_size,
            center[2],
        ]
        self.camera_backup["z_normal"] = [
            center[0],
            center[1],
            center[2] - distance_fudge_factor * box_mean_size,
        ]

    @property
    def empty(self) -> bool:
        """
        Check if the canvas is empty.
        """
        return not self.dims

    def home(self):
        """
        Reset the camera position.
        """
        self.move_camera(
            position=self.camera_backup["reset"], look_at=self.camera_backup["look_at"]
        )

    def camera_x_normal(self):
        """
        View scene along the X normal.
        """
        self._camera_normal(position=self.camera_backup["x_normal"].copy(), ind=0)

    def camera_y_normal(self):
        """
        View scene along the Y normal.
        """
        self._camera_normal(position=self.camera_backup["y_normal"].copy(), ind=1)

    def camera_z_normal(self):
        """
        View scene along the Z normal.
        """
        self._camera_normal(position=self.camera_backup["z_normal"].copy(), ind=2)

    def _camera_normal(self, position: tuple[float, float, float], ind: int):
        """
        Move camera to requested normal, and flip if current position is equal
        to the requested position.
        """
        if np.allclose(self.camera.position, position):
            position[ind] = 2.0 * self.camera_backup["center"][ind] - position[ind]
        self.move_camera(position=position)

    def move_camera(
        self,
        position: tuple[float, float, float],
        look_at: tuple[float, float, float] | None = None,
    ):
        """
        Move the camera to the supplied position.

        Parameters
        ----------
        position:
            The new camera position.
        """
        self.camera.position = position
        if look_at is None:
            look_at = self.camera_backup["center"]
        self.controls.target = look_at
        self.camera.lookAt(look_at)

    def toggle_outline(self):
        """
        Toggle the visibility of the outline box.
        """
        self.outline.visible = not self.outline.visible

    def toggle_axes3d(self):
        """
        Toggle the visibility of the RGB helper axes.
        """
        self.axes_3d.visible = not self.axes_3d.visible

    def add(self, obj: Any):
        """
        Add an object to the ``scene``.
        """
        self.scene.add(obj)

    def remove(self, obj):
        """
        Remove an object from the ``scene``.
        """
        self.scene.remove(obj)

    def _make_title(self):
        if self._title_text:
            html = (
                f'<div style="text-align: center; width: {self.figsize[0]}px">'
                f'{self._title_text}</div>'
            )
        else:
            html = None
        return ipw.HTML(html)

    @property
    def title(self) -> str:
        """
        Get or set the title of the plot.
        """
        return self._title_text

    @title.setter
    def title(self, text: str):
        self._title_text = text
        self._title = self._make_title()

    def bigger(self):
        self.figsize = (self.figsize * 1.2).astype(int)
        self._update_size()

    def smaller(self):
        self.figsize = (self.figsize / 1.2).astype(int)
        self._update_size()

    def _update_size(self):
        self.renderer.width = self.figsize[0]
        self.renderer.height = self.figsize[1]
        self.camera.aspect = self.figsize[0] / self.figsize[1]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from typing import Literal

from .. import backends
from ..utils import deprecated


@deprecated('Use ``linefigure`` instead.')
def figure1d(*args, style: Literal['line'] = 'line', **kwargs):
    """
    Create a figure to represent one-dimensional data from one or more graph node(s).
    By default, this will return a figure built from :class:`LineView` (see the
    documentation of this class for a list of available customization arguments).

    .. deprecated:: v24.04.0

    Parameters
    ----------
    style:
        The type of figure to create.

    Examples
    --------
    Create an input node and attach a figure1d as a view:

      >>> da = pp.data.data1d()
      >>> in_node = pp.Node(da)
      >>> fig = pp.figure1d(in_node)

    Visualize two data arrays on the same figure:

      >>> a = pp.data.data1d()
      >>> b = 3 * a
      >>> a_node = pp.Node(a)
      >>> b_node = pp.Node(b)
      >>> fig = pp.figure1d(a_node, b_node)

    With a customization argument to make the vertical scale logarithmic:

      >>> da = pp.data.data1d()
      >>> in_node = pp.Node(da)
      >>> fig = pp.figure1d(in_node, norm='log')
    """

    if style == 'line':
        from .lineview import LineView

        return backends.figure1d(LineView, *args, **kwargs)

    raise ValueError(f'Unsupported style={style} for figure1d.')


@deprecated('Use ``imagefigure`` instead.')
def figure2d(*args, style: Literal['image'] = 'image', **kwargs):
    """
    Create a figure to represent two-dimensional data from a graph node.
    By default, this will return a figure built from :class:`ImageView` (see the
    documentation of this class for a list of available customization arguments).

    .. deprecated:: v24.04.0

    Parameters
    ----------
    style:
        The type of figure to create.

    Examples
    --------
    Create an input node and attach a figure2d as a view:

      >>> da = pp.data.data2d()
      >>> in_node = pp.Node(da)
      >>> fig = pp.figure2d(in_node)

    With a customization argument to make the color scale logarithmic:

      >>> da = pp.data.data2d()
      >>> in_node = pp.Node(da)
      >>> fig = pp.figure2d(in_node, norm='log')
    """

    if style == 'image':
        from .imageview import ImageView

        return backends.figure2d(ImageView, *args, **kwargs)

    raise ValueError(f'Unsupported style={style} for figure2d.')


@deprecated('Use ``scatter3dfigure`` instead.')
def figure3d(*args, style: Literal['scatter'] = 'scatter', **kwargs):
    """
    Create a figure to represent three-dimensional data from a graph node.
    By default, this will return a figure built from :class:`FigScatter3d` (see the
    documentation of this class for a list of available customization arguments).

    .. deprecated:: v24.04.0

    Parameters
    ----------
    style:
        The type of figure to create.

    Examples
    --------
    Create an input node and attach a figure3d as a view:

      >>> da = pp.data.scatter()
      >>> in_node = pp.Node(da)
      >>> fig = pp.figure3d(in_node)

    With a customization argument to make the color scale logarithmic:

      >>> da = pp.data.scatter()
      >>> in_node = pp.Node(da)
      >>> fig = pp.figure3d(in_node, norm='log')
    """

    if style == 'scatter':
        from .scatter3dview import Scatter3dView

        return backends.figure3d(Scatter3dView, *args, **kwargs)

    raise ValueError(f'Unsupported style={style} for figure3d.')



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import scipp as sc

from plopp import Node
from plopp.data.testing import data_array, scatter
from plopp.graphics import scatter3dfigure
from plopp.widgets import ClippingPlanes


def test_add_remove_cuts():
    da = scatter()
    fig = scatter3dfigure(Node(da), x='x', y='y', z='z')
    clip = ClippingPlanes(fig)
    assert len(fig.artists) == 1
    clip.add_x_cut.click()
    assert len(fig.artists) == 2
    npoints_in_cutx = list(fig.artists.values())[-1]._data.shape[0]
    clip.add_y_cut.click()
    assert len(fig.artists) == 2
    npoints_in_cutxy = list(fig.artists.values())[-1]._data.shape[0]
    assert npoints_in_cutxy > npoints_in_cutx
    clip.add_z_cut.click()
    assert len(fig.artists) == 2
    npoints_in_cutxyz = list(fig.artists.values())[-1]._data.shape[0]
    assert npoints_in_cutxyz > npoints_in_cutxy
    clip.delete_cut.click()
    # If the tool is not displayed, the tab selected index does not update when a cut
    # is deleted, so we need to manually set it to the correct value
    clip.tabs.selected_index = 1
    assert list(fig.artists.values())[-1]._data.shape[0] == npoints_in_cutxy
    clip.delete_cut.click()
    # If the tool is not displayed, the tab selected index does not update when a cut
    # is deleted, so we need to manually set it to the correct value
    clip.tabs.selected_index = 0
    assert list(fig.artists.values())[-1]._data.shape[0] == npoints_in_cutx
    clip.delete_cut.click()
    assert len(fig.artists) == 1


def test_move_cut():
    da = scatter()
    fig = scatter3dfigure(Node(da), x='x', y='y', z='z')
    clip = ClippingPlanes(fig)
    clip.add_x_cut.click()
    xcut = clip.cuts[-1]
    assert xcut.outlines[0].position[0] == xcut.slider.value[0]
    assert xcut.outlines[1].position[0] == xcut.slider.value[1]
    pts = list(fig.artists.values())[-1]
    npoints = pts._data.shape[0]
    xcut.slider.value = [xcut.slider.min, xcut.slider.value[1]]
    assert xcut.outlines[0].position[0] == xcut.slider.value[0]
    assert xcut.outlines[1].position[0] == xcut.slider.value[1]
    clip.update_state()  # Need to manually update state due to debounce mechanism
    new_pts = list(fig.artists.values())[-1]
    assert npoints < new_pts._data.shape[0]


def test_operation_or():
    dim = 'pix'
    da = data_array(ndim=3).flatten(to=dim)
    fig = scatter3dfigure(Node(da), x='xx', y='yy', z='zz')
    clip = ClippingPlanes(fig)
    clip.cut_operation.value = 'OR'

    clip.add_x_cut.click()
    xcut = clip.cuts[-1]
    data_in_xcut = list(clip._nodes.values())[-1]['slice']()
    xrange = xcut.slider.value
    xsel = (da.coords['xx'] >= sc.scalar(xrange[0], unit='m')) & (
        da.coords['xx'] < sc.scalar(xrange[1], unit='m')
    )
    expected = da[xsel].flatten(to=dim)
    assert sc.identical(expected, data_in_xcut)

    clip.add_y_cut.click()
    ycut = clip.cuts[-1]
    data_in_xycut = list(clip._nodes.values())[-1]['slice']()
    yrange = ycut.slider.value
    ysel = (da.coords['yy'] >= sc.scalar(yrange[0], unit='m')) & (
        da.coords['yy'] < sc.scalar(yrange[1], unit='m')
    )
    expected = da[xsel | ysel].flatten(to=dim)
    assert sc.identical(expected, data_in_xycut)

    clip.add_z_cut.click()
    zcut = clip.cuts[-1]
    data_in_xyzcut = list(clip._nodes.values())[-1]['slice']()
    zrange = zcut.slider.value
    zsel = (da.coords['zz'] >= sc.scalar(zrange[0], unit='m')) & (
        da.coords['zz'] < sc.scalar(zrange[1], unit='m')
    )
    expected = da[xsel | ysel | zsel].flatten(to=dim)
    assert sc.identical(expected, data_in_xyzcut)


def test_operation_and():
    dim = 'pix'
    da = data_array(ndim=3).flatten(to=dim)
    fig = scatter3dfigure(Node(da), x='xx', y='yy', z='zz')
    clip = ClippingPlanes(fig)
    clip.cut_operation.value = 'AND'

    clip.add_x_cut.click()
    xcut = clip.cuts[-1]
    xrange = xcut.slider.value
    xsel = (da.coords['xx'] >= sc.scalar(xrange[0], unit='m')) & (
        da.coords['xx'] < sc.scalar(xrange[1], unit='m')
    )

    clip.add_y_cut.click()
    ycut = clip.cuts[-1]
    data_in_xycut = list(clip._nodes.values())[-1]['slice']()
    yrange = ycut.slider.value
    ysel = (da.coords['yy'] >= sc.scalar(yrange[0], unit='m')) & (
        da.coords['yy'] < sc.scalar(yrange[1], unit='m')
    )
    expected = da[xsel & ysel].flatten(to=dim)
    assert sc.identical(expected, data_in_xycut)

    clip.add_z_cut.click()
    zcut = clip.cuts[-1]
    data_in_xyzcut = list(clip._nodes.values())[-1]['slice']()
    zrange = zcut.slider.value
    zsel = (da.coords['zz'] >= sc.scalar(zrange[0], unit='m')) & (
        da.coords['zz'] < sc.scalar(zrange[1], unit='m')
    )
    expected = da[xsel & ysel & zsel].flatten(to=dim)
    assert sc.identical(expected, data_in_xyzcut)


def test_operation_xor():
    dim = 'pix'
    da = data_array(ndim=3).flatten(to=dim)
    fig = scatter3dfigure(Node(da), x='xx', y='yy', z='zz')
    clip = ClippingPlanes(fig)
    clip.cut_operation.value = 'XOR'

    clip.add_x_cut.click()
    xcut = clip.cuts[-1]
    xrange = xcut.slider.value
    xsel = (da.coords['xx'] >= sc.scalar(xrange[0], unit='m')) & (
        da.coords['xx'] < sc.scalar(xrange[1], unit='m')
    )

    clip.add_y_cut.click()
    ycut = clip.cuts[-1]
    data_in_xycut = list(clip._nodes.values())[-1]['slice']()
    yrange = ycut.slider.value
    ysel = (da.coords['yy'] >= sc.scalar(yrange[0], unit='m')) & (
        da.coords['yy'] < sc.scalar(yrange[1], unit='m')
    )
    expected = da[(xsel | ysel) & ~(xsel & ysel)].flatten(to=dim)
    assert sc.identical(expected, data_in_xycut)

    clip.add_z_cut.click()
    zcut = clip.cuts[-1]
    data_in_xyzcut = list(clip._nodes.values())[-1]['slice']()
    zrange = zcut.slider.value
    zsel = (da.coords['zz'] >= sc.scalar(zrange[0], unit='m')) & (
        da.coords['zz'] < sc.scalar(zrange[1], unit='m')
    )
    expected = da[
        (xsel | ysel | zsel) & ~(xsel & ysel) & ~(xsel & zsel) & ~(ysel & zsel)
    ].flatten(to=dim)
    assert sc.identical(expected, data_in_xyzcut)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from collections.abc import Sequence
from typing import Any

import scipp as sc

from ..core.utils import maybe_variable_to_number


def _vector_to_tuple(
    vector: sc.Variable | Sequence[sc.Variable] | Sequence[float],
) -> tuple[sc.Variable | float, ...]:
    if isinstance(vector, sc.Variable):
        return (vector.fields.x, vector.fields.y, vector.fields.z)
    else:
        return tuple(v for v in vector)


class Camera:
    """
    Camera configuration for three-dimensional plots.
    If values are provided as raw numbers instead of Scipp variables, their unit
    will be assumed to be the same as the unit of the spatial coordinates.

    Parameters
    ----------
    position:
        The position of the camera, as a :func:`scipp.vector` or a list of 3 numbers.
    look_at:
        The point the camera is looking at, as a :func:`scipp.vector` or a list of 3
        numbers.
    near:
        The distance to the near clipping plane (how close to the camera objects can be
        before they disappear), as a :func:`scipp.scalar` or a single number.
    far:
        The distance to the far clipping plane (how far from the camera objects can be
        before they disappear), as a :func:`scipp.scalar` or a single number.
    """

    def __init__(
        self,
        position: sc.Variable | Sequence[sc.Variable] | Sequence[float] | None = None,
        look_at: sc.Variable | Sequence[sc.Variable] | Sequence[float] | None = None,
        near: sc.Variable | float | None = None,
        far: sc.Variable | float | None = None,
    ):
        self._parsed_contents = None
        self._raw_contents = {}
        if position is not None:
            self._raw_contents['position'] = _vector_to_tuple(position)
        if look_at is not None:
            self._raw_contents['look_at'] = _vector_to_tuple(look_at)
        if near is not None:
            self._raw_contents['near'] = near
        if far is not None:
            self._raw_contents['far'] = far

    def get(self, key: str, default: Any | None = None) -> Any:
        """
        Attribute getter.
        If units have been set, the value will be converted to the specified units.
        Otherwise, the raw value will be returned.

        Parameters
        ----------
        key:
            The name of the attribute to get.
        default:
            The default value to return if the attribute is not set.
        """
        return (
            self._parsed_contents.get(key, default)
            if self._parsed_contents is not None
            else self._raw_contents.get(key, default)
        )

    def set_units(self, xunit: str, yunit: str, zunit: str):
        """
        Set the units of the camera attributes.
        This will convert the raw values to the specified units.

        Parameters
        ----------
        xunit:
            The unit of the x axis.
        yunit:
            The unit of the y axis.
        zunit:
            The unit of the z axis.
        """
        self._parsed_contents = {}
        for key in set(self._raw_contents) & {'position', 'look_at'}:
            self._parsed_contents[key] = tuple(
                maybe_variable_to_number(x, unit=u)
                for x, u in zip(
                    self._raw_contents[key], [xunit, yunit, zunit], strict=True
                )
            )

        for key in set(self._raw_contents) & {'near', 'far'}:
            if isinstance(self._raw_contents[key], sc.Variable):
                if not (xunit == yunit == zunit):
                    raise sc.UnitError(
                        'All axes must have the same unit when specifying a clipping '
                        f'plane with a unit. Units are: {xunit}, {yunit}, {zunit}.'
                    )
            self._parsed_contents[key] = maybe_variable_to_number(
                self._raw_contents[key], unit=xunit
            )

    def has_units(self) -> bool:
        """
        Check if units have been set.
        """
        return self._parsed_contents is not None

    @property
    def position(self) -> tuple[float, float, float] | None:
        """
        The position of the camera. If camera units have been set, the position returned
        will be converted to the specified units.
        """
        self.get('position')

    @property
    def look_at(self) -> tuple[float, float, float] | None:
        """
        The point the camera is looking at. If camera units have been set, the position
        returned will be converted to the specified units.
        """
        self.get('look_at')

    @property
    def near(self) -> float | None:
        """
        The distance to the near clipping plane (how close to the camera objects can be
        before they disappear).
        """
        self.get('near')

    @property
    def far(self) -> float | None:
        """
        The distance to the far clipping plane (how far from the camera objects can be
        before they disappear).
        """
        self.get('far')

    def __repr__(self):
        return (
            f'Camera(position={self.position}, look_at={self.look_at}, '
            f'near={self.near}, far={self.far})'
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import pytest
import scipp as sc

import plopp as pp


@pytest.mark.usefixtures('_use_ipympl')
def test_creation():
    da = pp.data.data3d()
    pp.inspector(da)


@pytest.mark.usefixtures('_use_ipympl')
def test_from_node():
    da = pp.data.data3d()
    pp.inspector(pp.Node(da))


@pytest.mark.usefixtures('_use_ipympl')
def test_multiple_inputs_raises():
    da = pp.data.data3d()
    with pytest.raises(TypeError, match='Cannot convert input of type'):
        pp.inspector({'a': da, 'b': 2.3 * da})


@pytest.mark.usefixtures('_use_ipympl')
def test_bad_number_of_dims_raises():
    with pytest.raises(
        ValueError,
        match='The inspector plot currently only works with three-dimensional data',
    ):
        pp.inspector(pp.data.data2d())
    with pytest.raises(
        ValueError,
        match='The inspector plot currently only works with three-dimensional data',
    ):
        pp.inspector(pp.data.data_array(ndim=4))


@pytest.mark.usefixtures('_use_ipympl')
def test_raises_ValueError_when_given_binned_data():
    da = sc.data.table_xyz(100).bin(x=10, y=20, z=30)
    with pytest.raises(ValueError, match='Cannot plot binned data'):
        pp.inspector(da, orientation='vertical')


@pytest.mark.usefixtures('_use_ipympl')
def test_line_creation():
    da = pp.data.data3d()
    ip = pp.inspector(da)
    fig2d = ip[0][0]
    fig1d = ip[0][1]
    # Activate the inspector tool
    fig2d.toolbar['inspect'].value = True
    assert not fig1d.canvas.dims
    assert len(fig1d.artists) == 0
    fig2d.toolbar['inspect']._tool.click(10, 10)
    assert fig1d.canvas.dims == {'x': da.dims[-1]}
    assert len(fig1d.artists) == 1
    fig2d.toolbar['inspect']._tool.click(20, 15)
    assert len(fig1d.artists) == 2


@pytest.mark.usefixtures('_use_ipympl')
def test_line_removal():
    da = pp.data.data3d()
    ip = pp.inspector(da)
    fig2d = ip[0][0]
    fig1d = ip[0][1]
    fig2d.toolbar['inspect'].value = True
    assert not fig1d.canvas.dims
    assert len(fig1d.artists) == 0
    fig2d.toolbar['inspect']._tool.click(10, 10)
    assert fig1d.canvas.dims == {'x': da.dims[-1]}
    assert len(fig1d.artists) == 1
    fig2d.toolbar['inspect']._tool.click(20, 15)
    assert len(fig1d.artists) == 2
    fig2d.toolbar['inspect']._tool.remove(1)
    assert len(fig1d.artists) == 1
    fig2d.toolbar['inspect']._tool.remove(0)
    assert len(fig1d.artists) == 0


@pytest.mark.usefixtures('_use_ipympl')
def test_operation():
    da = pp.data.data3d()
    ip_sum = pp.inspector(da, operation='sum')
    ip_mean = pp.inspector(da, operation='mean')
    assert ip_sum[0][0]._view.colormapper.vmax > ip_mean[0][0]._view.colormapper.vmax
    assert ip_sum[0][0]._view.colormapper.vmin < ip_mean[0][0]._view.colormapper.vmin



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import scipp as sc

from plopp.backends.matplotlib.canvas import Canvas
from plopp.backends.matplotlib.image import Image
from plopp.data.testing import data_array


def test_image_creation():
    da = data_array(ndim=2)
    image = Image(canvas=Canvas(), data=da)
    assert sc.identical(image._data, da)


def test_image_creation_binedges():
    da = data_array(ndim=2, binedges=True)
    image = Image(canvas=Canvas(), data=da)
    assert sc.identical(image._data, da)


def test_image_creation_masks():
    da = data_array(ndim=2, masks=True)
    image = Image(canvas=Canvas(), data=da)
    assert sc.identical(image._data, da)


def test_image_creation_ragged_coord():
    da = data_array(ndim=2, ragged=True)
    image = Image(canvas=Canvas(), data=da)
    assert image._dim_2d == ('x', 'xx')
    assert image._dim_1d == ('y', 'yy')
    assert sc.identical(image._data, da)


def test_image_update():
    da = data_array(ndim=2)
    image = Image(canvas=Canvas(), data=da)
    assert sc.identical(image._data, da)
    image.update(da * 2.5)
    assert sc.identical(image._data, da * 2.5)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)


from ..core import View


class BaseFig:
    """
    A Mixin class which is the base for all figures.
    """

    _view: View

    @property
    def canvas(self):
        return self._view.canvas

    @property
    def artists(self):
        return self._view.artists

    @property
    def dims(self):
        return self._view.dims

    @property
    def graph_nodes(self):
        return self._view.graph_nodes

    @property
    def id(self):
        return self._view.id

    def update(self, *args, **kwargs):
        return self._view.update(*args, **kwargs)

    def notify_view(self, *args, **kwargs):
        return self._view.notify_view(*args, **kwargs)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

import numpy as np
import scipp as sc

from plopp.backends.matplotlib.canvas import Canvas
from plopp.backends.matplotlib.scatter import Scatter
from plopp.data.testing import scatter as scatter_data


def test_scatter_creation():
    da = scatter_data()
    scat = Scatter(canvas=Canvas(), data=da)
    assert scat._unit == da.unit
    assert len(scat._scatter.get_offsets()) == len(da)
    x, y = scat._scatter.get_offsets().T
    assert np.allclose(x, da.coords['x'].values)
    assert np.allclose(y, da.coords['y'].values)
    assert not scat._mask.get_visible()


def test_scatter_with_mask():
    da = scatter_data()
    da.masks['mask'] = da.coords['x'] > sc.scalar(5, unit='m')
    scat = Scatter(canvas=Canvas(), data=da)
    assert scat._mask.get_visible()
    expected = da[da.masks['mask']]
    x, y = scat._mask.get_offsets().T
    assert np.allclose(x[~np.isnan(x)], expected.coords['x'].values)
    assert np.allclose(y[~np.isnan(y)], expected.coords['y'].values)


def test_scatter_update():
    da = scatter_data()
    scat = Scatter(canvas=Canvas(), data=da)
    x, y = scat._scatter.get_offsets().T
    assert np.allclose(x, da.coords['x'].values)
    assert np.allclose(y, da.coords['y'].values)
    new = da.copy()
    new.coords['x'] += sc.scalar(1.1, unit=da.coords['x'].unit)
    new.coords['y'] *= 2.3
    scat.update(new)
    x, y = scat._scatter.get_offsets().T
    assert np.allclose(x, new.coords['x'].values)
    assert np.allclose(y, new.coords['y'].values)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import uuid

import numpy as np
import scipp as sc
from matplotlib.lines import Line2D

from ...core.utils import merge_masks
from .canvas import Canvas
from .utils import make_legend, parse_dicts_in_kwargs


class Scatter:
    """ """

    def __init__(
        self,
        canvas: Canvas,
        data: sc.DataArray,
        x: str = 'x',
        y: str = 'y',
        size: str | None = None,
        number: int = 0,
        mask_color: str = 'black',
        cbar: bool = False,
        **kwargs,
    ):
        self._canvas = canvas
        self._ax = self._canvas.ax
        self._data = data
        self._x = x
        self._y = y
        self._size = size
        # Because all keyword arguments from the figure are forwarded to both the canvas
        # and the line, we need to remove the arguments that belong to the canvas.
        kwargs.pop('ax', None)

        scatter_kwargs = parse_dicts_in_kwargs(kwargs, name=data.name)

        self.label = data.name if not cbar else None
        self._unit = self._data.unit
        self._id = uuid.uuid4().hex

        markers = list(Line2D.markers.keys())
        default_plot_style = {
            'marker': markers[(number + 2) % len(markers)],
        }
        if not cbar:
            default_plot_style['color'] = f'C{number}'

        merged_kwargs = {**default_plot_style, **scatter_kwargs}
        if self._size is None:
            s = merged_kwargs.pop('s', None)
        else:
            s = self._data.coords[self._size].values

        self._scatter = self._ax.scatter(
            self._data.coords[self._x].values,
            self._data.coords[self._y].values,
            s=s,
            label=self.label,
            **merged_kwargs,
        )

        xmask = self._data.coords[self._x].values.copy()
        ymask = self._data.coords[self._y].values.copy()
        visible_mask = False
        if self._data.masks:
            not_one_mask = ~merge_masks(self._data.masks).values
            xmask[not_one_mask] = np.nan
            ymask[not_one_mask] = np.nan
            visible_mask = True
        self._mask = self._ax.scatter(
            xmask,
            ymask,
            s=s,
            marker=merged_kwargs['marker'],
            edgecolors=mask_color,
            facecolor="None",
            linewidth=3.0,
            zorder=self._scatter.get_zorder() + 1,
            visible=visible_mask,
        )

        if self._canvas._legend:
            leg_args = make_legend(self._canvas._legend)
            if np.shape(s) == np.shape(self._data.coords[self._x].values):
                handles, labels = self._scatter.legend_elements(prop="sizes")
                self._ax.legend(handles, labels, title="Sizes", **leg_args)
            if self.label:
                self._ax.legend(**leg_args)

    def update(self, new_values: sc.DataArray):
        """
        Update the x and y positions of the data points from new data.

        Parameters
        ----------
        new_values:
            New data to update the line values, masks, errorbars from.
        """
        self._data = new_values
        self._scatter.set_offsets(
            np.stack(
                [self._data.coords[self._x].values, self._data.coords[self._y].values],
                axis=1,
            )
        )
        if self._size is not None:
            self._scatter.set_sizes(self._data.coords[self._size].values)

    def remove(self):
        """
        Remove the scatter and mask artists from the canvas.
        """
        self._scatter.remove()
        self._mask.remove()

    def set_colors(self, rgba: np.ndarray):
        if self._scatter.get_array() is not None:
            self._scatter.set_array(None)
        self._scatter.set_facecolors(rgba)

    @property
    def data(self):
        """ """
        return self._data



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import numpy as np
import pytest
import scipp as sc

from plopp import Node
from plopp.data.testing import data_array
from plopp.graphics.lineview import LineView


def test_empty():
    fig = LineView()
    assert len(fig.artists) == 0


def test_update():
    fig = LineView()
    assert len(fig.artists) == 0
    da = data_array(ndim=1)
    fig.update(data1d=da)
    assert sc.identical(fig.artists['data1d']._data, da)


def test_update_not_1d_raises():
    fig = LineView()
    with pytest.raises(ValueError, match="Expected 1 dimension"):
        fig.update(data2d=data_array(ndim=2))
    with pytest.raises(ValueError, match="Expected 1 dimension"):
        fig.update(data3d=data_array(ndim=3))


def test_create_with_node():
    da = data_array(ndim=1)
    fig = LineView(Node(da))
    assert len(fig.artists) == 1
    [line] = fig.artists.values()
    assert sc.identical(line._data, da)
    assert line._error is None


def test_with_errorbars():
    da = data_array(ndim=1, variances=True)
    fig = LineView(Node(da))
    assert len(fig.artists) == 1
    [line] = fig.artists.values()
    assert line._error is not None
    fig = LineView(Node(da), errorbars=False)
    [line] = fig.artists.values()
    assert line._error is None


def test_with_binedges():
    da = data_array(ndim=1, binedges=True)
    fig = LineView(Node(da))
    assert len(fig.artists) == 1
    [line] = fig.artists.values()
    assert sc.identical(line._data, da)
    xdata = line._line.get_xdata()
    assert np.allclose(xdata, da.coords['xx'].values)
    assert len(xdata) == da.sizes['xx'] + 1


def test_log_norm():
    fig = LineView()
    assert fig.canvas.yscale == 'linear'
    fig = LineView(norm='log')
    assert fig.canvas.yscale == 'log'


def test_update_grows_limits():
    da = data_array(ndim=1)
    fig = LineView(Node(da))
    old_lims = fig.canvas.yrange
    [key] = fig.artists.keys()
    fig.update({key: da * 2.5})
    new_lims = fig.canvas.yrange
    assert new_lims[0] < old_lims[0]
    assert new_lims[1] > old_lims[1]


def test_update_does_shrink_limits_if_auto_mode():
    da = data_array(ndim=1)
    fig = LineView(Node(da), autoscale='auto')
    old_lims = fig.canvas.yrange
    [key] = fig.artists.keys()
    const = 0.5
    fig.update({key: da * const})
    new_lims = fig.canvas.yrange
    assert new_lims[0] == old_lims[0] * const
    assert new_lims[1] == old_lims[1] * const


def test_update_does_not_shrink_limits_if_grow_mode():
    da = data_array(ndim=1)
    fig = LineView(Node(da), autoscale='grow')
    old_lims = fig.canvas.yrange
    [key] = fig.artists.keys()
    fig.update({key: da * 0.5})
    new_lims = fig.canvas.yrange
    assert new_lims[0] == old_lims[0]
    assert new_lims[1] == old_lims[1]


def test_vmin():
    da = data_array(ndim=1)
    fig = LineView(Node(da), vmin=sc.scalar(-0.5, unit='m/s'))
    assert fig.canvas.ymin == -0.5


def test_vmin_unit_mismatch_raises():
    da = data_array(ndim=1)
    with pytest.raises(sc.UnitError):
        _ = LineView(Node(da), vmin=sc.scalar(-0.5, unit='m'))


def test_vmax():
    da = data_array(ndim=1)
    fig = LineView(Node(da), vmax=sc.scalar(0.68, unit='m/s'))
    assert fig.canvas.ymax == 0.68


def test_vmin_vmax():
    da = data_array(ndim=1)
    fig = LineView(
        Node(da),
        vmin=sc.scalar(-0.5, unit='m/s'),
        vmax=sc.scalar(0.68, unit='m/s'),
    )
    assert np.allclose(fig.canvas.yrange, [-0.5, 0.68])


def test_vmin_vmax_no_variable():
    da = data_array(ndim=1)
    fig = LineView(Node(da), vmin=-0.5, vmax=0.68)
    assert np.allclose(fig.canvas.yrange, [-0.5, 0.68])


def test_raises_for_new_data_with_incompatible_dimension():
    x = data_array(ndim=1)
    y = x.rename(xx='yy')
    with pytest.raises(KeyError):
        LineView(Node(x), Node(y))


def test_raises_for_new_data_with_incompatible_unit():
    a = data_array(ndim=1)
    b = a * a
    with pytest.raises(sc.UnitError):
        LineView(Node(a), Node(b))


def test_raises_for_new_data_with_incompatible_coord_unit():
    a = data_array(ndim=1)
    b = a.copy()
    b.coords['xx'] = a.coords['xx'] * a.coords['xx']
    with pytest.raises(sc.UnitError):
        LineView(Node(a), Node(b))


def test_converts_new_data_units():
    a = data_array(ndim=1, unit='m')
    b = data_array(ndim=1, unit='cm')
    anode = Node(a)
    bnode = Node(b)
    fig = LineView(anode, bnode)
    assert sc.identical(fig.artists[anode.id]._data, a)
    assert sc.identical(fig.artists[bnode.id]._data, b.to(unit='m'))


def test_converts_new_data_coordinate_units():
    a = data_array(ndim=1)
    b = data_array(ndim=1)
    b.coords['xx'].unit = 'cm'
    anode = Node(a)
    bnode = Node(b)
    fig = LineView(anode, bnode)
    assert sc.identical(fig.artists[anode.id]._data, a)
    c = b.copy()
    c.coords['xx'] = c.coords['xx'].to(unit='m')
    assert sc.identical(fig.artists[bnode.id]._data, c)


def test_converts_new_data_units_integers():
    a = sc.DataArray(
        data=sc.array(dims=['x'], values=[1, 2, 3, 4, 5], unit='m'),
        coords={'x': sc.arange('x', 5.0, unit='s')},
    )
    b = sc.DataArray(
        data=sc.array(dims=['x'], values=[10, 20, 30, 40, 50], unit='cm'),
        coords={'x': sc.arange('x', 5.0, unit='s')},
    )
    anode = Node(a)
    bnode = Node(b)
    fig = LineView(anode, bnode)
    assert sc.identical(fig.artists[anode.id]._data, a)
    assert sc.identical(fig.artists[bnode.id]._data, b.to(unit='m', dtype=float))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from typing import Literal

import scipp as sc

from .. import backends
from ..core.typing import FigureLike
from .colormapper import ColorMapper
from .graphicalview import GraphicalView


class ImageView(GraphicalView):
    """
    ImageView that makes a visual representation of two-dimensional data.
    It has a :class:`Canvas`, a :class:`ColorMapper` and a specialized ``update``
    function that generates :class:`Image` artists.

    Parameters
    ----------
    *nodes:
        The nodes that are attached to the view.
    cmap:
        The name of the colormap for the data
        (see https://matplotlib.org/stable/tutorials/colors/colormaps.html).
        In addition to the Matplotlib docs, if the name is just a single html color,
        a colormap with that single color will be used.
    mask_cmap:
        The name of the colormap for masked data.
    norm:
        Control the scaling on the vertical axis.
    vmin:
        Lower bound for the colorbar. If a number (without a unit) is supplied, it is
        assumed that the unit is the same as the data unit.
    vmax:
        Upper bound for the colorbar. If a number (without a unit) is supplied, it is
        assumed that the unit is the same as the data unit.
    autoscale:
        The behavior of the color range limits. If ``auto``, the limits automatically
        adjusts every time the data changes. If ``grow``, the limits are allowed to
        grow with time but they do not shrink.
    scale:
        Control the scaling of the horizontal axis. For example, specify
        ``scale={'tof': 'log'}`` if you want log-scale for the ``tof`` dimension.
    aspect:
        Aspect ratio for the axes.
    grid:
        Show grid if ``True``.
    cbar:
        Show colorbar if ``True``.
    title:
        The figure title.
    figsize:
        The width and height of the figure, in inches.
    ax:
        If supplied, use these axes to create the figure. If none are supplied, the
        canvas will create its own axes.
    cax:
        If supplied, use these axes for the colorbar. If none are supplied, and a
        colorbar is required, the canvas will create its own axes.
    format:
        Format of the figure displayed in the Jupyter notebook. If ``None``, a SVG is
        created as long as the number of markers in the figure is not too large. If too
        many markers are drawn, a PNG image is created instead.
    **kwargs:
        All other kwargs are forwarded to the Image artist.
    """

    def __init__(
        self,
        *nodes,
        cmap: str = 'viridis',
        mask_cmap: str = 'gray',
        norm: Literal['linear', 'log'] = 'linear',
        vmin: sc.Variable | float | None = None,
        vmax: sc.Variable | float | None = None,
        autoscale: Literal['auto', 'grow'] = 'auto',
        scale: dict[str, str] | None = None,
        aspect: Literal['auto', 'equal'] = 'auto',
        grid: bool = False,
        cbar: bool = True,
        title: str | None = None,
        figsize: tuple[float, float] | None = None,
        format: Literal['svg', 'png'] | None = None,
        **kwargs,
    ):
        super().__init__(*nodes)

        # Note: the order of dims matters here, reversing it would transpose the image
        self._dims = {'y': None, 'x': None}
        self._ndim = 2
        self._scale = {} if scale is None else scale
        self._kwargs = kwargs
        self._repr_format = format
        self.canvas = backends.canvas2d(
            cbar=cbar, aspect=aspect, grid=grid, title=title, figsize=figsize, **kwargs
        )
        self.colormapper = ColorMapper(
            cmap=cmap,
            cbar=cbar,
            mask_cmap=mask_cmap,
            norm=norm,
            vmin=vmin,
            vmax=vmax,
            autoscale=autoscale,
            canvas=self.canvas,
        )

        self.render()
        self.canvas.autoscale()
        self.canvas.finalize()

    def make_artist(self, new_values):
        return backends.image(canvas=self.canvas, data=new_values, **self._kwargs)


def imagefigure(*args, **kwargs) -> FigureLike:
    """
    Create a figure to represent two-dimensional data from one or more graph node(s).

    .. versionadded:: 24.04.0

    Examples
    --------
    Create an input node and attach an ``imagefigure`` as a view:

      >>> da = pp.data.data2d()
      >>> in_node = pp.Node(da)
      >>> fig = pp.imagefigure(in_node)

    With a customization argument to make the color scale logarithmic:

      >>> da = pp.data.data2d()
      >>> in_node = pp.Node(da)
      >>> fig = pp.imagefigure(in_node, norm='log')
    """

    return backends.figure2d(ImageView, *args, **kwargs)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)


def _make_figure(*args, **kwargs):
    from .utils import is_interactive_backend

    if is_interactive_backend():
        from .interactive import InteractiveFig

        return InteractiveFig(*args, **kwargs)
    else:
        from .static import StaticFig

        return StaticFig(*args, **kwargs)


class MatplotlibBackend:
    def is_interactive(self):
        """
        Returns ``True`` if the backend currently in use allows for interactive figures.
        """
        from .utils import is_interactive_backend

        return is_interactive_backend()

    def canvas2d(self, *args, **kwargs):
        """
        See :class:`canvas.Canvas` for details.
        """
        from .canvas import Canvas as CanvasMpl

        return CanvasMpl(*args, **kwargs)

    def line(self, *args, **kwargs):
        """
        See :class:`line.Line` for details.
        """
        from .line import Line as LineMpl

        return LineMpl(*args, **kwargs)

    def scatter(self, *args, **kwargs):
        """
        See :class:`scatter.Scatter` for details.
        """
        from .scatter import Scatter as ScatterMpl

        return ScatterMpl(*args, **kwargs)

    def image(self, *args, **kwargs):
        """
        See :class:`image.Image` for details.
        """
        from .image import Image as ImageMpl

        return ImageMpl(*args, **kwargs)

    def figure1d(self, *args, **kwargs):
        """
        See :class:`static.StaticFig` and :class:`interactive.InteractiveFig` for
        details.
        """
        return _make_figure(*args, **kwargs)

    def figure2d(self, *args, **kwargs):
        """
        See :class:`static.StaticFig` and :class:`interactive.InteractiveFig` for
        details.
        """
        return _make_figure(*args, **kwargs)

    def tiled(self, *args, **kwargs):
        """ """
        from .tiled import Tiled as TiledMpl

        return TiledMpl(*args, **kwargs)


__all__ = ["MatplotlibBackend"]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import pytest
import scipp as sc

from plopp import Node
from plopp.data.testing import scatter
from plopp.graphics.scatter3dview import Scatter3dView


def test_creation():
    da = scatter()
    fig = Scatter3dView(Node(da), x='x', y='y', z='z')
    assert len(fig.artists) == 1
    [key] = fig.artists.keys()
    assert sc.identical(fig.artists[key]._data, da)


def test_update():
    da = scatter()
    fig = Scatter3dView(Node(da), x='x', y='y', z='z')
    assert len(fig.artists) == 1
    [key] = fig.artists.keys()
    fig.update({key: da * 3.3})
    assert sc.identical(fig.artists[key]._data, da * 3.3)


def test_log_norm():
    da = scatter()
    fig = Scatter3dView(Node(da), x='x', y='y', z='z', norm='log')
    assert fig.colormapper.norm == 'log'


def test_raises_for_new_data_with_incompatible_coordinate():
    a = scatter()
    b = scatter()
    b.coords['t'] = b.coords.pop('x')
    with pytest.raises(KeyError):
        Scatter3dView(Node(a), Node(b), x='x', y='y', z='z')


def test_raises_for_new_data_with_incompatible_unit():
    a = scatter()
    b = a * a
    with pytest.raises(sc.UnitError):
        Scatter3dView(Node(a), Node(b), x='x', y='y', z='z')


def test_raises_for_new_data_with_incompatible_coord_unit():
    a = scatter()
    b = a.copy()
    b.coords['x'] = a.coords['x'] * a.coords['x']
    with pytest.raises(sc.UnitError):
        Scatter3dView(Node(a), Node(b), x='x', y='y', z='z')


def test_converts_new_data_units():
    a = scatter()
    a.unit = 'm'
    b = scatter()
    b.unit = 'cm'
    anode = Node(a)
    bnode = Node(b)
    fig = Scatter3dView(anode, bnode, x='x', y='y', z='z')
    assert sc.identical(fig.artists[anode.id]._data, a)
    assert sc.identical(fig.artists[bnode.id]._data, b.to(unit='m'))


def test_converts_new_data_coordinate_units():
    a = scatter()
    b = scatter()
    xcoord = b.coords['x'].copy()
    xcoord.unit = 'cm'
    b.coords['x'] = xcoord
    anode = Node(a)
    bnode = Node(b)
    fig = Scatter3dView(anode, bnode, x='x', y='y', z='z')
    assert sc.identical(fig.artists[anode.id]._data, a)
    c = b.copy()
    c.coords['x'] = c.coords['x'].to(unit='m')
    assert sc.identical(fig.artists[bnode.id]._data, c)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

import uuid
from functools import partial
from typing import Literal

import scipp as sc

from ..core.typing import FigureLike, PlottableMulti
from .common import check_not_binned, from_compatible_lib, input_to_nodes


def _preprocess_scatter(
    obj: PlottableMulti,
    x: str,
    y: str,
    size: str | None,
    name: str | None = None,
):
    da = from_compatible_lib(obj)
    check_not_binned(da)

    cnames = [x, y]
    if size is not None:
        cnames.append(size)
    coords = {k: da.coords[k] for k in cnames}
    out = sc.DataArray(data=da.data, masks=da.masks, coords=coords)
    if out.ndim != 1:
        out = out.flatten(to=uuid.uuid4().hex)
    if name is not None:
        out.name = name
    return out


def scatter(
    obj: PlottableMulti,
    *,
    x: str = 'x',
    y: str = 'y',
    size: str | float | None = None,
    figsize: tuple[float, float] | None = None,
    norm: Literal['linear', 'log'] = 'linear',
    title: str | None = None,
    vmin: sc.Variable | float = None,
    vmax: sc.Variable | float = None,
    cbar: bool = False,
    cmap: str = 'viridis',
    **kwargs,
) -> FigureLike:
    """
    Make a two-dimensional scatter plot.

    .. versionadded:: 24.04.0

    Parameters
    ----------
    obj:
        The object to be plotted.
    x:
        The name of the coordinate that is to be used for the X positions.
    y:
        The name of the coordinate that is to be used for the Y positions.
    size:
        The size of the marker. If a float is supplied, all markers will have the same
        size. If a string is supplied, it will be the name of the coordinate that is to
        be used for the size of the markers.
    figsize:
        The width and height of the figure, in inches.
    norm:
        Set to ``'log'`` for a logarithmic colorscale (only applicable if ``cbar`` is
        ``True``).
    title:
        The figure title.
    vmin:
        Lower bound for the colorscale for (only applicable if ``cbar`` is ``True``).
    vmax:
        Upper bound for the colorscale for (only applicable if ``cbar`` is ``True``).
    cbar:
        Show colorbar if ``True``. If ``cbar`` is ``True``, the marker will be colored
        using the data values in the supplied data array.
    cmap:
        The colormap to be used for the colorscale.
    **kwargs:
        All other kwargs are forwarded the underlying plotting library.
    """
    from ..graphics import scatterfigure

    nodes = input_to_nodes(
        obj, processor=partial(_preprocess_scatter, x=x, y=y, size=size)
    )

    return scatterfigure(
        *nodes,
        x=x,
        y=y,
        size=size,
        figsize=figsize,
        norm=norm,
        title=title,
        vmin=vmin,
        vmax=vmax,
        cmap=cmap,
        cbar=cbar,
        **kwargs,
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from ipywidgets import HBox, VBox, Widget


class Bar:
    """
    A simple mixin to provide add and remove helper methods for HBox/VBox containers.
    """

    def __len__(self):
        return len(self.children)

    def add(self, obj: Widget):
        """
        Append a widget to the list of children.
        """
        self.children = [*self.children, obj]

    def remove(self, obj: Widget):
        """
        Remove a widget from the list of children.
        """
        children = list(self.children)
        children.remove(obj)
        self.children = children


class VBar(VBox, Bar):
    """
    Vertical bar container.
    """

    def __getitem__(self, ind):
        if isinstance(ind, int):
            return self.children[ind]
        elif isinstance(ind, slice):
            return VBar(self.children[ind])


class HBar(HBox, Bar):
    """
    Horizontal bar container.
    """

    def __getitem__(self, ind):
        if isinstance(ind, int):
            return self.children[ind]
        elif isinstance(ind, slice):
            return HBar(self.children[ind])


class Box(VBar):
    """
    Container widget that accepts a list of items. For each item in the list, if the
    item is itself a list, it will be made into a horizontal row of the underlying
    items, if not, the item will span then entire row.
    Finally, all the rows will be placed inside a vertical box container.

    Parameters
    ----------

    widgets:
        List of widgets to place in the box.
    """

    def __init__(self, widgets):
        children = [
            HBar(view) if isinstance(view, list | tuple) else view for view in widgets
        ]
        super().__init__(children)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import os
import tempfile

import pytest

from plopp.backends.plotly.canvas import Canvas

pytest.importorskip("plotly")


def test_creation():
    title = 'My canvas'
    canvas = Canvas(title=title)
    assert canvas.title == title


def test_logx():
    canvas = Canvas()
    canvas.logx()
    assert canvas.xscale == 'log'
    canvas.logx()
    assert canvas.xscale == 'linear'


def test_logy():
    canvas = Canvas()
    canvas.logy()
    assert canvas.yscale == 'log'
    canvas.logy()
    assert canvas.yscale == 'linear'


@pytest.mark.parametrize('ext', ['jpg', 'png', 'pdf', 'svg', 'html'])
def test_save_to_disk(ext):
    canvas = Canvas()
    with tempfile.TemporaryDirectory() as path:
        fname = os.path.join(path, f'plopp_fig.{ext}')
        canvas.save(filename=fname)
        assert os.path.isfile(fname)


def test_save_to_disk_with_bad_extension_raises():
    canvas = Canvas()
    with pytest.raises(ValueError, match='txt'):
        canvas.save(filename='plopp_fig.txt')



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import pytest
import scipp as sc

from plopp import Node
from plopp.data.testing import data_array
from plopp.plotting.superplot import superplot


@pytest.mark.usefixtures('_use_ipympl')
def test_creation():
    da = data_array(ndim=2)
    sp = superplot(da, keep='xx')
    assert len(sp.right_bar[0]._lines) == 0


@pytest.mark.usefixtures('_use_ipympl')
def test_from_node():
    da = data_array(ndim=2)
    superplot(Node(da))


@pytest.mark.usefixtures('_use_ipympl')
def test_save_line():
    da = data_array(ndim=2)
    sp = superplot(da, keep='xx')
    tool = sp.right_bar[0]
    assert len(tool._lines) == 0
    tool.save_line()
    assert len(tool._lines) == 1
    line = list(tool._lines.values())[-1]
    assert sc.identical(line['line']._data, da['yy', 0])
    assert len(tool.container.children) == 1

    slider = sp.bottom_bar[0]
    slider.controls['yy']['slider'].value = 5
    tool.save_line()
    assert len(tool._lines) == 2
    line = list(tool._lines.values())[-1]
    assert sc.identical(line['line']._data, da['yy', 5])
    assert len(tool.container.children) == 2


@pytest.mark.usefixtures('_use_ipympl')
def test_remove_line():
    da = data_array(ndim=2)
    sp = superplot(da, keep='xx')
    tool = sp.right_bar[0]
    assert len(tool._lines) == 0
    tool.save_line()
    slider = sp.bottom_bar[0]
    slider.controls['yy']['slider'].value = 5
    tool.save_line()
    slider.controls['yy']['slider'].value = 15
    tool.save_line()
    assert len(tool._lines) == 3
    assert len(tool.container.children) == 3

    keys = list(tool._lines.keys())
    first_line = keys[0]
    last_line = keys[2]

    tool.remove_line(change=None, line_id=first_line)
    assert first_line not in tool._lines
    assert len(tool.container.children) == 2

    tool.remove_line(change=None, line_id=last_line)
    assert last_line not in tool._lines
    assert len(tool.container.children) == 1


@pytest.mark.usefixtures('_use_ipympl')
def test_change_line_color():
    da = data_array(ndim=2)
    sp = superplot(da, keep='xx')
    tool = sp.right_bar[0]
    tool.save_line()
    line_id = list(tool._lines.keys())[-1]
    tool.change_line_color(change={'new': '#000000'}, line_id=line_id)
    assert tool._lines[line_id]['line'].color == '#000000'


@pytest.mark.usefixtures('_use_ipympl')
def test_raises_ValueError_when_given_binned_data():
    da = sc.data.table_xyz(100).bin(x=10, y=20)
    with pytest.raises(ValueError, match='Cannot plot binned data'):
        superplot(da, keep='x')



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)


import numpy as np
import pythreejs as p3
from matplotlib import ticker
from scipp import Variable

from ...core.utils import value_to_string


def _get_delta(x: Variable, axis: int) -> float:
    """
    Compute the difference between two bounds along a given axis.
    """
    return (x[axis][1] - x[axis][0]).value


def _get_offsets(
    limits: tuple[Variable, Variable, Variable], axis: int, ind: int
) -> np.ndarray:
    """
    Compute offsets for 3 dimensions, along the edges of the box.
    """
    offsets = np.array([limits[i][ind].value for i in range(3)])
    offsets[axis] = 0
    return offsets


def _make_geometry(limits: tuple[Variable, Variable, Variable]) -> p3.EdgesGeometry:
    """
    Make a geometry to represent the edges of a cubic box.
    """
    return p3.EdgesGeometry(
        p3.BoxBufferGeometry(
            width=_get_delta(limits, axis=0),
            height=_get_delta(limits, axis=1),
            depth=_get_delta(limits, axis=2),
        )
    )


def _make_sprite(
    string: str,
    position: tuple[float, float, float],
    color: str = "black",
    size: float = 1.0,
) -> p3.Sprite:
    """
    Make a text-based sprite for axis tick.
    """
    sm = p3.SpriteMaterial(
        map=p3.TextTexture(string=string, color=color, size=300, squareTexture=True),
        transparent=True,
    )
    return p3.Sprite(material=sm, position=position, scale=[size, size, size])


class Outline(p3.Group):
    """
    Create an object that draws a rectangular outline, given some limits for the XYZ
    dimensions. Along the lower edges of the cube are added some tick labels and axes
    labels, according to the dimension extents, dimension names, and units given in
    the limits.

    Parameters
    ----------
    limits:
        A tuple of variables, each of length 2, which contain the lower and upper bounds
        for the outline. Each variable also has a dimension, which is to be used as the
        dimension for that direction, as well as a unit, which will be used to label
        the corresponding axis.
    tick_size:
        A number to scale the tick size.
    """

    def __init__(
        self,
        limits: tuple[Variable, Variable, Variable],
        tick_size: float | None = None,
    ):
        center = [var.mean().value for var in limits]
        if tick_size is None:
            tick_size = 0.05 * np.mean([_get_delta(limits, axis=i) for i in range(3)])

        self.box = p3.LineSegments(
            geometry=_make_geometry(limits),
            material=p3.LineBasicMaterial(color='#000000'),
            position=center,
        )

        self.ticks = self._make_ticks(limits=limits, tick_size=tick_size)
        self.labels = self._make_labels(
            limits=limits, center=center, tick_size=tick_size
        )

        super().__init__()
        for obj in (self.box, self.ticks, self.labels):
            self.add(obj)

    def _make_ticks(
        self, limits: tuple[Variable, Variable, Variable], tick_size: float
    ) -> p3.Group:
        """
        Create tick labels on outline edges
        """
        ticks_group = p3.Group()
        iden = np.identity(3, dtype=np.float32)
        ticker_ = ticker.MaxNLocator(5)
        for axis in range(3):
            ticks = ticker_.tick_values(limits[axis][0].value, limits[axis][1].value)
            for tick in ticks:
                if limits[axis][0].value <= tick <= limits[axis][1].value:
                    tick_pos = iden[axis] * tick + _get_offsets(limits, axis, 0)
                    ticks_group.add(
                        _make_sprite(
                            string=value_to_string(tick, precision=1),
                            position=tick_pos.tolist(),
                            size=tick_size,
                        )
                    )
        return ticks_group

    def _make_labels(
        self,
        limits: tuple[Variable, Variable, Variable],
        center: list[float],
        tick_size: float,
    ) -> p3.Group:
        """
        Create axes labels (coord dimension and unit) on outline edges
        """
        labels_group = p3.Group()
        for axis in range(3):
            axis_label = f'{limits[axis].dim} [{limits[axis].unit}]'
            # Offset labels 5% beyond axis ticks to reduce overlap
            delta = 0.05
            labels_group.add(
                _make_sprite(
                    string=axis_label,
                    position=(
                        np.roll([1, 0, 0], axis) * center[axis]
                        + (1.0 + delta) * _get_offsets(limits, axis, 0)
                        - delta * _get_offsets(limits, axis, 1)
                    ).tolist(),
                    size=tick_size * 0.3 * len(axis_label),
                )
            )

        return labels_group



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import numpy as np
import pytest
import scipp as sc

from plopp.backends.plotly.canvas import Canvas
from plopp.backends.plotly.line import Line
from plopp.data.testing import data_array

pytest.importorskip("plotly")


def test_line_creation():
    da = data_array(ndim=1, unit='K')
    line = Line(canvas=Canvas(), data=da)
    assert line._unit == 'K'
    assert line._dim == 'xx'
    assert len(line._line.x) == da.sizes['xx']
    assert np.allclose(line._line.x, da.coords['xx'].values)
    assert np.allclose(line._line.y, da.values)
    assert line._error is None
    assert not line._mask.visible


def test_line_creation_bin_edges():
    da = data_array(ndim=1, binedges=True)
    line = Line(canvas=Canvas(), data=da)
    assert len(line._line.x) == da.sizes['xx'] + 1


def test_line_with_errorbars():
    da = data_array(ndim=1, variances=True)
    line = Line(canvas=Canvas(), data=da)
    assert np.allclose(line._error.error_y['array'], sc.stddevs(da.data).values)


def test_line_with_bin_edges_and_errorbars():
    da = data_array(ndim=1, binedges=True, variances=True)
    line = Line(canvas=Canvas(), data=da)
    assert np.allclose(line._error.x, sc.midpoints(da.coords['xx']).values)


def test_line_hide_errorbars():
    da = data_array(ndim=1, variances=True)
    line = Line(canvas=Canvas(), data=da, errorbars=False)
    assert line._error is None


def test_line_with_mask():
    da = data_array(ndim=1, masks=True)
    line = Line(canvas=Canvas(), data=da)
    assert line._mask.visible


def test_line_with_mask_and_binedges():
    da = data_array(ndim=1, binedges=True, masks=True)
    line = Line(canvas=Canvas(), data=da)
    assert line._mask.visible


def test_line_with_two_masks():
    da = data_array(ndim=1, masks=True)
    da.masks['two'] = da.coords['xx'] > sc.scalar(25, unit='m')
    line = Line(canvas=Canvas(), data=da)
    expected = da.data[da.masks['mask'] | da.masks['two']].values
    y = line._mask.y
    assert np.allclose(y[~np.isnan(y)], expected)


def test_line_update():
    da = data_array(ndim=1)
    line = Line(canvas=Canvas(), data=da)
    assert np.allclose(line._line.x, da.coords['xx'].values)
    assert np.allclose(line._line.y, da.values)
    line.update(da * 2.5)
    assert np.allclose(line._line.x, da.coords['xx'].values)
    assert np.allclose(line._line.y, da.values * 2.5)


def test_line_update_with_errorbars():
    da = data_array(ndim=1, variances=True)
    line = Line(canvas=Canvas(), data=da)
    assert np.allclose(line._line.y, da.values)
    assert np.allclose(line._error.error_y['array'], sc.stddevs(da.data).values)
    new_values = da * 3.3
    new_values.variances = da.variances
    line.update(new_values)
    assert np.allclose(line._line.y, da.values * 3.3)
    assert np.allclose(line._error.error_y['array'], sc.stddevs(da.data).values)
    new_values = 1.0 * da
    new_values.variances = da.variances * 4.0
    line.update(new_values)
    assert np.allclose(line._line.y, da.values)
    assert np.allclose(line._error.error_y['array'], sc.stddevs(da.data).values * 2.0)


def test_line_datetime_binedges_with_errorbars():
    t = np.arange(
        np.datetime64('2017-03-16T20:58:17'), np.datetime64('2017-03-16T21:15:17'), 20
    )
    time = sc.array(dims=['time'], values=t)
    v = np.random.rand(time.sizes['time'] - 1)
    da = sc.DataArray(
        data=sc.array(dims=['time'], values=10 * v, variances=v), coords={'time': time}
    )
    xint = t.astype(int)
    xmid = (0.5 * (xint[1:] + xint[:-1])).astype(int)
    expected = np.array(xmid, dtype=t.dtype)
    line = Line(canvas=Canvas(), data=da)
    # Note that allclose does not work on datetime dtypes
    assert np.allclose(line._error.x.astype(int), expected.astype(int))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import pytest
import scipp as sc

from plopp import Node
from plopp.data.testing import data_array, dataset
from plopp.plotting.slicer import Slicer


@pytest.mark.usefixtures('_use_ipympl')
def test_creation_keep_two_dims():
    da = data_array(ndim=3)
    sl = Slicer(da, keep=['xx', 'yy'])
    assert sl.slider.value == {'zz': 0}
    assert sl.slider.controls['zz']['slider'].max == da.sizes['zz'] - 1
    assert sc.identical(sl.slice_nodes[0].request_data(), da['zz', 0])


@pytest.mark.usefixtures('_use_ipympl')
def test_creation_keep_one_dim():
    da = data_array(ndim=3)
    sl = Slicer(da, keep=['xx'])
    assert sl.slider.value == {'zz': 0, 'yy': 0}
    assert sl.slider.controls['yy']['slider'].max == da.sizes['yy'] - 1
    assert sl.slider.controls['zz']['slider'].max == da.sizes['zz'] - 1
    assert sc.identical(sl.slice_nodes[0].request_data(), da['yy', 0]['zz', 0])


@pytest.mark.usefixtures('_use_ipympl')
def test_update_keep_two_dims():
    da = data_array(ndim=3)
    sl = Slicer(da, keep=['xx', 'yy'])
    assert sl.slider.value == {'zz': 0}
    assert sc.identical(sl.slice_nodes[0].request_data(), da['zz', 0])
    sl.slider.controls['zz']['slider'].value = 5
    assert sl.slider.value == {'zz': 5}
    assert sc.identical(sl.slice_nodes[0].request_data(), da['zz', 5])


@pytest.mark.usefixtures('_use_ipympl')
def test_update_keep_one_dim():
    da = data_array(ndim=3)
    sl = Slicer(da, keep=['xx'])
    assert sl.slider.value == {'zz': 0, 'yy': 0}
    assert sc.identical(sl.slice_nodes[0].request_data(), da['yy', 0]['zz', 0])
    sl.slider.controls['yy']['slider'].value = 5
    assert sl.slider.value == {'zz': 0, 'yy': 5}
    assert sc.identical(sl.slice_nodes[0].request_data(), da['yy', 5]['zz', 0])
    sl.slider.controls['zz']['slider'].value = 8
    assert sl.slider.value == {'zz': 8, 'yy': 5}
    assert sc.identical(sl.slice_nodes[0].request_data(), da['yy', 5]['zz', 8])


@pytest.mark.usefixtures('_use_ipympl')
def test_with_dataset():
    ds = dataset(ndim=2)
    sl = Slicer(ds, keep=['xx'])
    nodes = list(sl.figure.graph_nodes.values())
    sl.slider.controls['yy']['slider'].value = 5
    assert sc.identical(nodes[0].request_data(), ds['a']['yy', 5])
    assert sc.identical(nodes[1].request_data(), ds['b']['yy', 5])


@pytest.mark.usefixtures('_use_ipympl')
def test_with_data_group():
    da = data_array(ndim=2)
    dg = sc.DataGroup(a=da, b=da * 2.5)
    sl = Slicer(dg, keep=['xx'])
    nodes = list(sl.figure.graph_nodes.values())
    sl.slider.controls['yy']['slider'].value = 5
    assert sc.identical(nodes[0].request_data(), dg['a']['yy', 5])
    assert sc.identical(nodes[1].request_data(), dg['b']['yy', 5])


@pytest.mark.usefixtures('_use_ipympl')
def test_with_dict_of_data_arrays():
    a = data_array(ndim=2)
    b = data_array(ndim=2) * 2.5
    sl = Slicer({'a': a, 'b': b}, keep=['xx'])
    nodes = list(sl.figure.graph_nodes.values())
    sl.slider.controls['yy']['slider'].value = 5
    assert sc.identical(nodes[0].request_data(), a['yy', 5])
    assert sc.identical(nodes[1].request_data(), b['yy', 5])


@pytest.mark.usefixtures('_use_ipympl')
def test_with_data_arrays_same_shape_different_coord():
    a = data_array(ndim=2)
    b = data_array(ndim=2) * 2.5
    b.coords['xx'] *= 1.5
    Slicer({'a': a, 'b': b}, keep=['xx'])


@pytest.mark.usefixtures('_use_ipympl')
def test_with_data_arrays_different_shape_along_keep_dim():
    a = data_array(ndim=2)
    b = data_array(ndim=2) * 2.5
    Slicer({'a': a, 'b': b['xx', :10]}, keep=['xx'])


@pytest.mark.usefixtures('_use_ipympl')
def test_with_data_arrays_different_shape_along_non_keep_dim_raises():
    a = data_array(ndim=2)
    b = data_array(ndim=2) * 2.5
    with pytest.raises(
        ValueError, match='Slicer plot: all inputs must have the same sizes'
    ):
        Slicer({'a': a, 'b': b['yy', :10]}, keep=['xx'])


@pytest.mark.usefixtures('_use_ipympl')
def test_raises_ValueError_when_given_binned_data():
    da = sc.data.table_xyz(100).bin(x=10, y=20)
    with pytest.raises(ValueError, match='Cannot plot binned data'):
        Slicer(da, keep=['xx'])


@pytest.mark.usefixtures('_use_ipympl')
@pytest.mark.parametrize('ndim', [2, 3])
def test_from_node(ndim):
    da = data_array(ndim=ndim)
    Slicer(Node(da))


@pytest.mark.usefixtures('_use_ipympl')
def test_mixing_raw_data_and_nodes():
    a = data_array(ndim=2)
    b = 6.7 * a
    Slicer({'a': Node(a), 'b': Node(b)})
    Slicer({'a': a, 'b': Node(b)})
    Slicer({'a': Node(a), 'b': b})


@pytest.mark.usefixtures('_use_ipympl')
def test_raises_when_requested_keep_dims_do_not_exist():
    da = data_array(ndim=3)
    with pytest.raises(
        ValueError, match='Slicer plot: one or more of the requested dims to be kept'
    ):
        Slicer(da, keep=['time'])


@pytest.mark.usefixtures('_use_ipympl')
def test_raises_when_number_of_keep_dims_requested_is_bad():
    da = data_array(ndim=4)
    with pytest.raises(
        ValueError, match='Slicer plot: the number of dims to be kept must be 1 or 2'
    ):
        Slicer(da, keep=['xx', 'yy', 'zz'])
    with pytest.raises(
        ValueError, match='Slicer plot: the list of dims to be kept cannot be empty'
    ):
        Slicer(da, keep=[])


@pytest.mark.usefixtures('_use_ipympl')
def test_autoscale_fixed():
    da = sc.DataArray(
        data=sc.arange('x', 5 * 10 * 20).fold(dim='x', sizes={'z': 20, 'y': 10, 'x': 5})
    )
    sl = Slicer(da, keep=['y', 'x'], autoscale='fixed')
    assert sl.figure._view.colormapper.vmin == 0
    assert sl.figure._view.colormapper.vmax == 5 * 10 * 20 - 1
    sl.slider.controls['z']['slider'].value = 5
    assert sl.figure._view.colormapper.vmin == 0
    assert sl.figure._view.colormapper.vmax == 5 * 10 * 20 - 1



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import pytest
from scipp import identical

from plopp.data.testing import data_array
from plopp.widgets import RangeSliceWidget, SliceWidget, slice_dims


@pytest.mark.parametrize("widget", [SliceWidget, RangeSliceWidget])
def test_slice_creation(widget):
    da = data_array(ndim=3)
    sw = widget(da, dims=['yy', 'xx'])
    assert sw._slider_dims == ['yy', 'xx']
    assert sw.controls['xx']['slider'].min == 0
    assert sw.controls['xx']['slider'].max == da.sizes['xx'] - 1
    assert sw.controls['xx']['slider'].description == 'xx'
    assert sw.controls['yy']['slider'].min == 0
    assert sw.controls['yy']['slider'].max == da.sizes['yy'] - 1
    assert sw.controls['yy']['slider'].description == 'yy'


def test_slice_value_property():
    da = data_array(ndim=3)
    sw = SliceWidget(da, dims=['yy', 'xx'])
    sw.controls['xx']['slider'].value = 10
    sw.controls['yy']['slider'].value = 15
    assert sw.value == {'xx': 10, 'yy': 15}


def test_slice_label_updates():
    da = data_array(ndim=3)
    da.coords['xx'] *= 1.1
    da.coords['yy'] *= 3.3
    sw = SliceWidget(da, dims=['yy', 'xx'])
    assert sw.controls['xx']['label'].value == '0.0 [m]'
    sw.controls['xx']['slider'].value = 10
    assert sw.controls['xx']['label'].value == '11.0 [m]'
    assert sw.controls['yy']['label'].value == '0.0 [m]'
    sw.controls['yy']['slider'].value = 15
    assert sw.controls['yy']['label'].value == '49.5 [m]'


def test_slice_dims():
    da = data_array(ndim=3)
    slices = {'xx': 8, 'yy': 7}
    expected = da['xx', slices['xx']]['yy', slices['yy']]
    assert identical(slice_dims().func(da, slices=slices), expected)


def test_range_slice_dims():
    da = data_array(ndim=3)
    slices = {'xx': (8, 9), 'yy': (7, 10)}
    expected = da['xx', slice(*slices['xx'])]['yy', slice(*slices['yy'])]
    assert identical(slice_dims().func(da, slices=slices), expected)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from plopp import Node, View, show_graph


class SimpleView(View):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.value = None

    def notify_view(self, message):
        self.value = message


def has_edge(graph, node1, node2):
    """
    Check if graph contains a given edge.
    """
    tail_name = graph._quote_edge(node1)
    head_name = graph._quote_edge(node2)
    for key in graph.body:
        if (
            graph._edge(tail=tail_name, head=head_name, attr='').strip()
            == key.split(' [label')[0].strip()
        ):
            return True
    return False


def test_two_nodes_parent_child():
    # [ a ]
    #   |
    # [ b ]
    a = Node(5)
    b = Node(lambda x: x - 2, a)
    g = show_graph(a)
    assert has_edge(g, a.id, b.id)


def test_two_children():
    #    [ a ]
    #     / \
    #    /   \
    # [ b ] [ c ]
    a = Node(5)
    b = Node(lambda x: x - 2, a)
    c = Node(lambda x: x + 2, a)
    g = show_graph(a)
    assert has_edge(g, a.id, b.id)
    assert has_edge(g, a.id, c.id)


def test_two_parents():
    # [ a ] [ b ]
    #    \   /
    #     \ /
    #    [ c ]
    a = Node(5)
    b = Node(9)
    c = Node(lambda x, y: x + y, a, b)
    g = show_graph(a)
    assert has_edge(g, a.id, c.id)
    assert has_edge(g, b.id, c.id)


def test_two_parents_two_children():
    # [ a ] [ b ]
    #    \   / |
    #     \ /  |
    #    [ c ] |
    #       \  |
    #        \/
    #      [ d ]
    a = Node(5)
    b = Node(9)
    c = Node(lambda x, y: x + y, a, b)
    d = Node(lambda x, y: x - y, c, b)
    g = show_graph(a)
    assert has_edge(g, a.id, c.id)
    assert has_edge(g, b.id, c.id)
    assert has_edge(g, c.id, d.id)
    assert has_edge(g, b.id, d.id)


def test_two_grandchildren_have_common_parent():
    #       [ a ]
    #      /     \
    #     /       \
    #    /         \
    # [ b ] [ d ] [ c ]
    #    \   / \   /
    #     \ /   \ /
    #    [ e ] [ f ]
    a = Node(0)
    b = Node(lambda x: x + 1, a)
    c = Node(lambda x: x + 2, a)
    d = Node(4)
    e = Node(lambda x, y: x + y, b, d)
    f = Node(lambda x, y: x + y, c, d)
    g = show_graph(a)
    assert has_edge(g, a.id, b.id)
    assert has_edge(g, a.id, c.id)
    assert has_edge(g, b.id, e.id)
    assert has_edge(g, d.id, e.id)
    assert has_edge(g, d.id, f.id)
    assert has_edge(g, c.id, f.id)


def test_graph_with_views():
    #    [ a ]
    #     / \
    #    /   \
    # [ b ] [ View ]
    #   |
    #   |
    # [ View ]
    a = Node(5)
    b = Node(lambda x: x - 2, x=a)
    av = SimpleView(a)
    bv = SimpleView(b)
    g = show_graph(a)
    assert has_edge(g, a.id, b.id)
    assert has_edge(g, a.id, av.id)
    assert has_edge(g, b.id, bv.id)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from .figure import Figure
from .utils import fig_to_bytes


def _make_png_repr(fig):
    return {'image/png': fig_to_bytes(fig, form='png')}


def _make_svg_repr(fig):
    return {'image/svg+xml': fig_to_bytes(fig, form='svg').decode()}


REPR_MAP = {'png': _make_png_repr, 'svg': _make_svg_repr}


def get_repr_maker(form=None, npoints=0):
    """
    Get the function used to create the mimebundle representation of a figure,
    based on the format requested and the number of points in the figure.
    """
    if form is None:
        form = 'png' if (npoints > 20_000) else 'svg'
    return REPR_MAP[form.lower()]


class StaticFig(Figure):
    """
    Create a static Matplotlib figure.
    The output will be either svg or png, depending on the number of drawn onto the
    canvas.
    """

    def __init__(self, View, *args, **kwargs):
        self.__init_figure__(View, *args, **kwargs)

    def _repr_mimebundle_(self, include=None, exclude=None) -> dict:
        """
        Mimebundle display representation for jupyter notebooks.
        """
        str_repr = str(self.fig)
        out = {'text/plain': str_repr[:-1] + f', {len(self.artists)} artists)'}
        if self._view._repr_format is not None:
            repr_maker = get_repr_maker(form=self._view._repr_format)
        else:
            npoints = sum(len(line.get_xdata()) for line in self._view.canvas.ax.lines)
            repr_maker = get_repr_maker(npoints=npoints)
        out.update(repr_maker(self._view.canvas.fig))
        return out

    def to_widget(self):
        """
        Convert the Matplotlib figure to an image widget.
        """
        return self._view.canvas.to_image()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from plopp import Node
from plopp.data.testing import scatter
from plopp.graphics.scatter3dview import Scatter3dView
from plopp.widgets import TriCutTool


def test_show_hide_cuts():
    da = scatter()
    fig = Scatter3dView(Node(da), x='x', y='y', z='z')
    tri = TriCutTool(fig)
    assert len(fig.artists) == 1
    tri.cut_x.button.value = True
    assert len(fig.artists) == 2
    tri.cut_y.button.value = True
    assert len(fig.artists) == 3
    tri.cut_z.button.value = True
    assert len(fig.artists) == 4
    tri.cut_x.button.value = False
    assert len(fig.artists) == 3
    tri.cut_y.button.value = False
    assert len(fig.artists) == 2
    tri.cut_z.button.value = False
    assert len(fig.artists) == 1


def test_move_cut():
    da = scatter()
    fig = Scatter3dView(Node(da), x='x', y='y', z='z')
    tri = TriCutTool(fig)
    tri.cut_x.button.value = True
    assert tri.cut_x.outline.position[0] == tri.cut_x.slider.value
    pts = list(fig.artists.values())[-1]
    npoints = pts.geometry.attributes['position'].array.shape[0]
    tri.cut_x.slider.value = -5.0
    assert tri.cut_x.outline.position[0] == tri.cut_x.slider.value
    new_pts = list(fig.artists.values())[-1]
    assert npoints != new_pts.geometry.attributes['position'].array.shape[0]


def test_cut_thickness():
    da = scatter()
    fig = Scatter3dView(Node(da), x='x', y='y', z='z')
    tri = TriCutTool(fig)
    tri.cut_x.button.value = True
    pts = list(fig.artists.values())[-1]
    npoints = pts.geometry.attributes['position'].array.shape[0]
    tri.cut_x.plus_minus.plus()
    new_pts = list(fig.artists.values())[-1]
    assert npoints < new_pts.geometry.attributes['position'].array.shape[0]
    tri.cut_x.plus_minus.minus()
    new_pts = list(fig.artists.values())[-1]
    assert npoints == new_pts.geometry.attributes['position'].array.shape[0]
    tri.cut_x.plus_minus.minus()
    new_pts = list(fig.artists.values())[-1]
    assert npoints > new_pts.geometry.attributes['position'].array.shape[0]


def test_empty_cut():
    da = scatter()
    fig = Scatter3dView(Node(da), x='x', y='y', z='z')
    tri = TriCutTool(fig)
    assert len(fig.artists) == 1
    tri.cut_y.button.value = True
    assert len(fig.artists) == 2
    tri.cut_y.slider.value = 34.0  # This cut should contain no points
    assert len(fig.artists) == 1



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
# ruff: noqa: E402, F401

import importlib.metadata

try:
    __version__ = importlib.metadata.version(__package__ or __name__)
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

from .backends.manager import BackendManager

backends = BackendManager()

from . import data
from .core import Node, View, node, show_graph, widget_node
from .graphics import (
    Camera,
    figure1d,
    figure2d,
    figure3d,
    imagefigure,
    linefigure,
    scatter3dfigure,
    scatterfigure,
    tiled,
)
from .plotting import inspector, plot, scatter, scatter3d, slicer, superplot, xyplot

del importlib


def show() -> None:
    """
    A function to display all the currently opened figures (note that this only applies
    to the figures created via the Matplotlib backend).
    See https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.show.html for more
    details.
    """
    import matplotlib.pyplot as plt

    plt.show()


__all__ = [
    'Camera',
    'Node',
    'View',
    'backends',
    'data',
    'figure1d',
    'figure2d',
    'figure3d',
    'imagefigure',
    'inspector',
    'linefigure',
    'node',
    'plot',
    'scatter',
    'scatterfigure',
    'scatter3d',
    'scatter3dfigure',
    'show',
    'show_graph',
    'slicer',
    'superplot',
    'tiled',
    'widget_node',
    'xyplot',
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)


class PythreejsBackend:
    def is_interactive(self):
        """
        Returns ``True`` if the backend currently in use allows for interactive figures.
        """
        return True

    def canvas3d(self, *args, **kwargs):
        """
        See :class:`canvas.Canvas` for details.
        """
        from .canvas import Canvas as CanvasP3js

        return CanvasP3js(*args, **kwargs)

    def point_cloud(self, *args, **kwargs):
        """
        See :class:`point_cloud.PointCloud` for details.
        """
        from .point_cloud import PointCloud as PointCloudP3js

        return PointCloudP3js(*args, **kwargs)

    def figure3d(self, *args, **kwargs):
        """
        See :class:`figure.Figure` for details.
        """
        from .figure import Figure as FigP3js

        return FigP3js(*args, **kwargs)


__all__ = ["PythreejsBackend"]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from .graph import show_graph
from .helpers import node, widget_node
from .node import Node
from .view import View

__all__ = ['Node', 'View', 'node', 'show_graph', 'widget_node']



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import numpy as np
import pytest
import scipp as sc

import plopp as pp
from plopp.data.testing import scatter


def test_scatter3d_from_pos():
    da = scatter()
    pp.scatter3d(da, pos='position')


def test_scatter3d_from_xyz():
    da = scatter()
    pp.scatter3d(da, x='x', y='y', z='z')


def test_scatter3d_from_xyz_using_defaults():
    da = scatter()
    pp.scatter3d(da)


def test_scatter3d_dimensions_are_flattened():
    nx = 12
    ny = 12
    x = sc.linspace(dim='x', start=-10.0, stop=10.0, num=nx, unit='m')
    y = sc.linspace(dim='y', start=-10.0, stop=10.0, num=ny, unit='m')
    da = sc.DataArray(
        data=sc.array(dims=['x', 'y'], values=np.random.rand(nx, ny)),
        coords={'position': sc.spatial.as_vectors(x, y, 0.0 * sc.units.m)},
    )
    p = pp.scatter3d(da, pos="position")
    assert next(iter(p.artists.values())).data.ndim == 1
    nz = 12
    z = sc.linspace(dim='z', start=-10.0, stop=10.0, num=nz, unit='m')
    da = sc.DataArray(
        data=sc.array(dims=['x', 'y', 'z'], values=np.random.rand(nx, ny, nz)),
        coords={'position': sc.spatial.as_vectors(x, y, z)},
    )
    p = pp.scatter3d(da, pos="position")
    assert next(iter(p.artists.values())).data.ndim == 1


def test_scatter3d_can_plot_scalar_data():
    da = sc.DataArray(
        data=sc.scalar(1.2), coords={'position': sc.vector(value=[1, 2, 3])}
    )
    p = pp.scatter3d(da, pos='position')
    assert next(iter(p.artists.values())).data.ndim == 1


def test_raises_ValueError_when_given_binned_data():
    nx = 10
    da = sc.data.table_xyz(100).bin(x=nx)
    x = sc.linspace(dim='x', start=-10.0, stop=10.0, num=nx, unit='m')
    da.coords['position'] = sc.spatial.as_vectors(x, x, x)
    with pytest.raises(ValueError, match='Cannot plot binned data'):
        pp.scatter3d(da, pos='position')


def test_raises_ValueError_when_given_ax_kwarg():
    da = scatter()
    with pytest.raises(ValueError, match='Keyword "ax" detected'):
        pp.scatter3d(da, x='x', y='y', z='z', ax='dummy')


def test_scatter3d_from_xarray():
    import xarray as xr

    N = 200
    data = np.random.random(N)
    x = np.random.normal(size=N)
    y = np.random.normal(size=N)
    z = np.random.normal(size=N)
    da = xr.DataArray(
        data,
        coords={'x': ('pixel', x), 'y': ('pixel', y), 'z': ('pixel', z)},
        dims=['pixel'],
    )
    pp.scatter3d(da, x='x', y='y', z='z')


def test_scatter3d_dict_of_inputs():
    da1 = scatter()
    da2 = scatter()
    da2.coords['x'] += sc.scalar(100, unit='m')
    fig = pp.scatter3d({'a': da1, 'b': da2})
    assert len(fig.artists) == 2


def test_scatter3d_from_node():
    pp.scatter3d(pp.Node(scatter()))


def test_scatter3d_from_multiple_nodes():
    a = scatter()
    b = scatter()
    b.coords['x'] += sc.scalar(100, unit='m')
    pp.scatter3d({'a': pp.Node(a), 'b': pp.Node(b)})


def test_scatter3d_mixing_raw_data_and_nodes():
    a = scatter()
    b = scatter()
    b.coords['x'] += sc.scalar(100, unit='m')
    pp.scatter3d({'a': a, 'b': pp.Node(b)})
    pp.scatter3d({'a': pp.Node(a), 'b': b})



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from typing import Literal

import scipp as sc

from .. import backends
from ..core import View
from ..core.typing import FigureLike
from ..core.utils import make_compatible
from .camera import Camera
from .colormapper import ColorMapper


class Scatter3dView(View):
    """
    View that makes a visual representation of three-dimensional scatter data.
    It has a :class:`Canvas`, a :class:`ColorMapper` and a specialized ``update``
    function that generates :class:`PointCloud` artists.

    Parameters
    ----------
    *nodes:
        The nodes that are attached to the view.
    x:
        The name of the coordinate to use for the X positions of the scatter points.
    y:
        The name of the coordinate to use for the Y positions of the scatter points.
    z:
        The name of the coordinate to use for the Z positions of the scatter points.
    cmap:
        The name of the colormap for the data
        (see https://matplotlib.org/stable/tutorials/colors/colormaps.html).
        In addition to the Matplotlib docs, if the name is just a single html color,
        a colormap with that single color will be used.
    mask_cmap:
        The name of the colormap for masked data.
    norm:
        Control the scaling on the vertical axis.
    vmin:
        Lower bound for the colorbar. If a number (without a unit) is supplied, it is
        assumed that the unit is the same as the data unit.
    vmax:
        Upper bound for the colorbar. If a number (without a unit) is supplied, it is
        assumed that the unit is the same as the data unit.
    title:
        The figure title.
    figsize:
        The width and height of the figure, in pixels.
    camera:
        Initial camera configuration (position, target).
    **kwargs:
        All other kwargs are forwarded to the PointCloud artist.
    """

    def __init__(
        self,
        *nodes,
        x: str = 'x',
        y: str = 'y',
        z: str = 'z',
        cmap: str = 'viridis',
        mask_cmap: str = 'gray',
        norm: Literal['linear', 'log'] = 'linear',
        vmin: sc.Variable | float | None = None,
        vmax: sc.Variable | float | None = None,
        figsize: tuple[int, int] = (600, 400),
        title: str | None = None,
        camera: Camera | None = None,
        **kwargs,
    ):
        super().__init__(*nodes)

        self._x = x
        self._y = y
        self._z = z
        self._kwargs = kwargs

        self.canvas = backends.canvas3d(figsize=figsize, title=title, camera=camera)
        self.colormapper = ColorMapper(
            cmap=cmap,
            mask_cmap=mask_cmap,
            norm=norm,
            vmin=vmin,
            vmax=vmax,
            nan_color="#f0f0f0",
            figsize=self.canvas.figsize,
        )

        self._original_artists = [n.id for n in nodes]
        self.render()

    def update(self, *args, **kwargs):
        """
        Update the view with new point clouds by either supplying a dictionary of
        new data or by keyword arguments.
        """
        new = dict(*args, **kwargs)
        mapping = {'x': self._x, 'y': self._y, 'z': self._z}
        for key, new_values in new.items():
            if self.canvas.empty:
                self.canvas.set_axes(
                    dims=mapping,
                    units={
                        x: new_values.coords[dim].unit for x, dim in mapping.items()
                    },
                    dtypes={
                        x: new_values.coords[dim].dtype for x, dim in mapping.items()
                    },
                )
                self.colormapper.unit = new_values.unit
            else:
                new_values.data = make_compatible(
                    new_values.data, unit=self.colormapper.unit
                )
                for xyz, dim in mapping.items():
                    new_values.coords[dim] = new_values.coords[dim].to(
                        unit=self.canvas.units[xyz], copy=False
                    )

            if key not in self.artists:
                pts = backends.point_cloud(
                    data=new_values, x=self._x, y=self._y, z=self._z, **self._kwargs
                )
                self.artists[key] = pts
                self.colormapper[key] = pts
                self.canvas.add(pts.points)
                if key in self._original_artists:
                    self.canvas.make_outline(limits=self.get_limits())

            self.artists[key].update(new_values=new_values)
        self.colormapper.update(**new)

    def get_limits(self) -> tuple[sc.Variable, sc.Variable, sc.Variable]:
        """
        Get global limits for all the point clouds in the scene.
        """
        xmin = None
        xmax = None
        ymin = None
        ymax = None
        zmin = None
        zmax = None
        for child in self.artists.values():
            xlims, ylims, zlims = child.get_limits()
            if xmin is None or xlims[0] < xmin:
                xmin = xlims[0]
            if xmax is None or xlims[1] > xmax:
                xmax = xlims[1]
            if ymin is None or ylims[0] < ymin:
                ymin = ylims[0]
            if ymax is None or ylims[1] > ymax:
                ymax = ylims[1]
            if zmin is None or zlims[0] < zmin:
                zmin = zlims[0]
            if zmax is None or zlims[1] > zmax:
                zmax = zlims[1]
        return (
            sc.concat([xmin, xmax], dim=self._x),
            sc.concat([ymin, ymax], dim=self._y),
            sc.concat([zmin, zmax], dim=self._z),
        )

    def set_opacity(self, alpha: float):
        """
        Update the opacity of the original children (not the cuts).

        Parameters
        ----------
        alpha:
            The opacity value, between 0 and 1.
        """
        for name in self._original_artists:
            self.artists[name].opacity = alpha

    def remove(self, key: str):
        """
        Remove an object from the scene.

        Parameters
        ----------
        key:
            The id of the object to be removed.
        """
        self.canvas.remove(self.artists[key].points)
        del self.artists[key]


def scatter3dfigure(*args, **kwargs) -> FigureLike:
    """
    Create a figure to represent three-dimensional data from one or more graph node(s).

    .. versionadded:: 24.04.0

    Examples
    --------
    Create an input node and attach a ``scatter3dfigure`` as a view:

      >>> da = pp.data.scatter()
      >>> in_node = pp.Node(da)
      >>> fig = pp.scatter3dfigure(in_node)

    With a customization argument to make the color scale logarithmic:

      >>> da = pp.data.scatter()
      >>> in_node = pp.Node(da)
      >>> fig = pp.scatter3dfigure(in_node, norm='log')
    """

    return backends.figure3d(Scatter3dView, *args, **kwargs)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from plopp.widgets import Checkboxes


def test_checkboxes_creation():
    desc = 'My Checkboxes'
    cbx = Checkboxes(['a', 'b', 'c'], description=desc)
    assert 'a' in cbx.checkboxes
    assert 'b' in cbx.checkboxes
    assert 'c' in cbx.checkboxes
    assert cbx.description.value == desc
    assert cbx.checkboxes['a'].value
    assert cbx.checkboxes['b'].value
    assert cbx.checkboxes['c'].value
    assert cbx.toggle_all_button.value


def test_checkboxes_initial_value():
    cbx = Checkboxes(['a', 'b', 'c'], value=False)
    assert not cbx.checkboxes['a'].value
    assert not cbx.checkboxes['b'].value
    assert not cbx.checkboxes['c'].value
    assert not cbx.toggle_all_button.value


def test_checkboxes_value_property():
    cbx = Checkboxes(['a', 'b', 'c'])
    cbx.checkboxes['b'].value = False
    assert cbx.value == {'a': True, 'b': False, 'c': True}
